# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from inspect import getdoc
from typing import List, Set, Any, Optional, Callable, Dict
from warnings import catch_warnings, simplefilter
from functools import wraps
from copy import deepcopy


from .error import ConfigError, LeadershipError, ValueErrorWarning
from .i18n import _
from .setting import ConfigBag, OptionBag, owners, groups, undefined, \
                     FORBIDDEN_SET_PROPERTIES, SPECIAL_PROPERTIES
from .config import KernelConfig, KernelGroupConfig, KernelMetaConfig, KernelMixConfig
from .option import RegexpOption, OptionDescription, ChoiceOption, Leadership
from .todict import TiramisuDict
from .autolib import Calculation


TIRAMISU_VERSION = 4


class TiramisuHelp:
    _tmpl_help = '    {0}\t{1}'

    def help(self,
             _display: bool=True) -> List[str]:
        def display(doc=''):
            if _display: # pragma: no cover
                print(doc)
        all_modules = dir(self)
        modules = []
        max_len = 0
        force = False
        for module_name in all_modules:
            if module_name in ['forcepermissive', 'unrestraint', 'nowarnings']:
                force = True
                max_len = max(max_len, len('forcepermissive'))
            elif module_name != 'help' and not module_name.startswith('_'):
                modules.append(module_name)
                max_len = max(max_len, len(module_name))
        modules.sort()

        display(_(getdoc(self)))
        display()
        if force:
            display(_('Settings:'))
            display(self._tmpl_help.format('forcepermissive',
                                           _('Access to option without verifying permissive '
                                             'properties'),
                                           ).expandtabs(max_len + 10))
            display(self._tmpl_help.format('unrestraint',
                                           _('Access to option without property restriction')
                                           ).expandtabs(max_len + 10))
            display(self._tmpl_help.format('nowarnings',
                                           _('Do not warnings during validation')
                                           ).expandtabs(max_len + 10))
            display()
        if isinstance(self, TiramisuDispatcherOption):
            doc = _(getdoc(self.__call__))
            display(_('Call: {}').format(doc))
            display()
        display(_('Commands:'))
        for module_name in modules:
            module = getattr(self, module_name)
            doc = _(getdoc(module))
            display(self._tmpl_help.format(module_name, doc).expandtabs(max_len + 10))
        display()

    def __dir__(self):
        if '_registers' in super().__dir__():
            return list(self._registers.keys())
        return super().__dir__()


class CommonTiramisu(TiramisuHelp):
    _validate_properties = True

    def _get_options_bag(self,
                         follower_not_apply_requires: bool,
                         ) -> OptionBag:
        try:
            options_bag = self._config_bag.context.get_sub_option_bag(self._config_bag,
                                                                      self._path,
                                                                      self._index,
                                                                      self._validate_properties,
                                                                      follower_not_apply_requires=follower_not_apply_requires,
                                                                      )
        except AssertionError as err:
            raise ConfigError(str(err))
        except Exception as err:
            raise err
        return options_bag


def option_type(typ):
    if not isinstance(typ, list):
        types = [typ]
    else:
        types = typ

    def wrapper(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            self = args[0]
            config_bag = self._config_bag
            if self._config_bag.context.impl_type == 'group' and 'group' in types:
                options_bag = [OptionBag(None,
                                         None,
                                         self._config_bag,
                                         path=self._path,
                                         )]
                kwargs['is_group'] = True
                return func(self, options_bag, *args[1:], **kwargs)
            options_bag = self._get_options_bag('with_index' not in types)
            option = options_bag[-1].option
            if option.impl_is_optiondescription() and 'optiondescription' in types or \
                    option.impl_is_optiondescription() and option.impl_is_leadership() and 'leadership' in types or \
                    not option.impl_is_optiondescription() and (
                        option.impl_is_symlinkoption() and 'symlink' in types or \
                        not option.impl_is_symlinkoption() and (
                            'option' in types or \
                            option.impl_is_leader() and 'leader' in types or \
                            option.impl_is_follower() and 'follower' in types or \
                            isinstance(option, ChoiceOption) and 'choice' in types)):
                if not option.impl_is_optiondescription() and \
                        not option.impl_is_symlinkoption() and \
                        option.impl_is_follower():
                    # default is "without_index"
                    if 'with_index' not in types and 'with_or_without_index' not in types and \
                            self._index is not None:
                        msg = _('please do not specify index '
                                f'({self.__class__.__name__}.{func.__name__})')
                        raise ConfigError(_(msg))
                    if self._index is None and 'with_index' in types:
                        msg = _('please specify index with a follower option '
                                f'({self.__class__.__name__}.{func.__name__})')
                        raise ConfigError(msg)
                return func(self, options_bag, *args[1:], **kwargs)
            msg = _('please specify a valid sub function '
                    f'({self.__class__.__name__}.{func.__name__})')
            raise ConfigError(msg)
        wrapped.func = func
        return wrapped
    return wrapper


class CommonTiramisuOption(CommonTiramisu):
    _validate_properties = False

    def __init__(self,
                 path: str,
                 index: Optional[int],
                 config_bag: ConfigBag,
                 ) -> None:
        self._path = path
        self._index = index
        self._config_bag = config_bag

    def __getattr__(self, subfunc):
        raise ConfigError(_(f'please specify a valid sub function ({self.__class__.__name__}.{subfunc})'))


class _TiramisuOptionWalk:
    def _list(self,
              root_option_bag,
              type,
              group_type,
              recursive,
              ):
        assert type in ('all', 'option', 'optiondescription'), \
                _('unknown list type {}').format(type)
        assert group_type is None or isinstance(group_type, groups.GroupType), \
                _("unknown group_type: {0}").format(group_type)
        options = []
        types = {'option': ['option'],
                 'optiondescription': ['optiondescription'],
                 'all': ['option', 'optiondescription']}[type]
        for option_bag in self._config_bag.context.walk(root_option_bag,
                                                        recursive=recursive,
                                                        types=types,
                                                        group_type=group_type,
                                                        flatten_leadership=True,
                                                        ):
            if isinstance(option_bag, dict):
                for opts_bag in option_bag.values():
                    if isinstance(opts_bag, OptionBag):
                        options.append(TiramisuOption(opts_bag.path,
                                                      opts_bag.index,
                                                      self._config_bag,
                                                      ))
                    else:
                        for opt_bag in opts_bag:
                            options.append(TiramisuOption(opt_bag.path,
                                                          opt_bag.index,
                                                          self._config_bag,
                                                          ))
            else:
                options.append(TiramisuOption(option_bag.path,
                                              option_bag.index,
                                              self._config_bag,
                                              ))
        return options


class _TiramisuOptionOptionDescription:
    """Manage option"""
    _validate_properties = False

    @option_type(['optiondescription', 'option', 'with_or_without_index', 'symlink'])
    def get(self, options_bag: List[OptionBag]):
        """Get Tiramisu option"""
        option_bag = options_bag[-1]
        return option_bag.option

    @option_type(['optiondescription'])
    def isleadership(self, options_bag: List[OptionBag]):
        """Test if option is a leader or a follower"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_is_leadership()

    @option_type(['optiondescription', 'option', 'with_or_without_index'])
    def doc(self, options_bag: List[OptionBag]):
        """Get option document"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_get_display_name()

    @option_type(['optiondescription', 'option', 'with_or_without_index'])
    def description(self, options_bag: List[OptionBag]):
        """Get option description"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_get_information('doc', None)

    @option_type(['optiondescription', 'option', 'symlink', 'with_or_without_index'])
    def name(self, options_bag: List[OptionBag]) -> str:
        """Get option name"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_getname()

    @option_type(['optiondescription', 'option', 'with_or_without_index', 'symlink'])
    def path(self, options_bag: List[OptionBag]) -> str:
        """Get option path"""
        option_bag = options_bag[-1]
        return option_bag.path

    @option_type(['optiondescription', 'option', 'symlink', 'with_or_without_index'])
    def has_dependency(self,
                       options_bag: List[OptionBag],
                       self_is_dep=True,
                       ) -> bool:
        """Test if option has dependency"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_has_dependency(self_is_dep)

    @option_type(['optiondescription', 'option'])
    def dependencies(self, options_bag: List[OptionBag]):
        """Get dependencies from this option"""
        option_bag = options_bag[-1]
        options = []
        for option in option_bag.option.get_dependencies(self._config_bag.context):
            options.append(TiramisuOption(option().impl_getpath(),
                                          None,
                                          self._config_bag,
                                          ))
        return options

    @option_type(['optiondescription', 'option', 'with_or_without_index', 'symlink'])
    def isoptiondescription(self, options_bag: List[OptionBag]):
        """Test if option is an optiondescription"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_is_optiondescription()

    @option_type(['optiondescription', 'option', 'with_or_without_index', 'symlink'])
    def properties(self,
                   options_bag: List[OptionBag],
                   only_raises=False,
                   uncalculated=False,
                   ):
        """Get properties for an option"""
        option_bag = options_bag[-1]
        settings = self._config_bag.context.get_settings()
        if uncalculated:
            return settings.getproperties(option_bag,
                                          uncalculated=True,
                                          )
        if not only_raises:
            return settings.getproperties(option_bag,
                                          apply_requires=False,
                                          )
        return settings.calc_raises_properties(option_bag,
                                               apply_requires=False,
                                               uncalculated=uncalculated,
                                               )


class _TiramisuOptionOption(_TiramisuOptionOptionDescription):
    """Manage option"""
    @option_type(['option', 'symlink', 'with_or_without_index'])
    def ismulti(self, options_bag: List[OptionBag]):
        """Test if option could have multi value"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_is_multi()

    @option_type(['option', 'symlink', 'with_or_without_index'])
    def issubmulti(self, options_bag: List[OptionBag]):
        """Test if option could have submulti value"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_is_submulti()

    @option_type(['option', 'with_or_without_index'])
    def isleader(self, options_bag: List[OptionBag]):
        """Test if option is a leader"""
        return options_bag[-1].option.impl_is_leader()

    @option_type(['option', 'with_or_without_index', 'symlink'])
    def isfollower(self, options_bag: List[OptionBag]):
        """Test if option is a follower"""
        return options_bag[-1].option.impl_is_follower()

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def isdynamic(self, options_bag: List[OptionBag]):
        """Test if option is a dynamic optiondescription"""
        return options_bag[-1].option.impl_is_dynsymlinkoption()

    @option_type(['option', 'symlink', 'with_or_without_index'])
    def issymlinkoption(self, options_bag: List[OptionBag]) -> bool:
        """Test if option is a symlink option"""
        return options_bag[-1].option.impl_is_symlinkoption()

    @option_type(['option', 'with_or_without_index', 'symlink'])
    def default(self, options_bag: List[OptionBag]):
        """Get default value for an option (not for optiondescription)"""
        return options_bag[-1].option.impl_getdefault()

    @option_type(['option', 'with_or_without_index', 'symlink'])
    def defaultmulti(self, options_bag: List[OptionBag]):
        """Get default value when added a value for a multi option (not for optiondescription)"""
        if not options_bag[-1].option.impl_is_multi():
            raise ConfigError(_('only multi value has defaultmulti'))
        return options_bag[-1].option.impl_getdefault_multi()

    @option_type(['option', 'optiondescription', 'symlink', 'with_or_without_index'])
    def type(self, options_bag: List[OptionBag]):
        """Get de option type"""
        option_bag = options_bag[-1]
        if option_bag.option.impl_is_optiondescription():
            return 'optiondescription'
        return option_bag.option.get_type()

    @option_type(['option', 'with_or_without_index'])
    def pattern(self, options_bag: List[OptionBag]) -> str:
        """Get the option pattern"""
        option_bag = options_bag[-1]
        option = option_bag.option
        type = option.get_type()
        if isinstance(option, RegexpOption):
            return option._regexp.pattern
        if type == _('integer'):
            # FIXME negative too!
            return r'^[0-9]+$'
        if type == _('domain name'):
            return option.impl_get_extra('_domain_re').pattern
        if type in [_('ip'), _('network'), _('netmask')]:
            #FIXME only from 0.0.0.0 to 255.255.255.255
            return r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'

    @option_type(['option', 'leadership'])
    def leader(self, options_bag: List[OptionBag]):
        """Get the leader option for a follower option"""
        option = options_bag[-1].option
        leadership = options_bag[-1].option
        if not isinstance(option, Leadership):
            leadership = leadership.impl_get_leadership()
        path = leadership.get_leader().impl_getpath()
        return TiramisuOption(path,
                              None,
                              self._config_bag,
                              )

    @option_type(['option', 'with_or_without_index'])
    def index(self, options_bag: List[OptionBag]):
        """Get then index of option"""
        return options_bag[-1].index


class TiramisuOptionOption(CommonTiramisuOption):
    def __call__(self,
                 name: str,
                 index: Optional[int]=None,
                 ) -> 'TiramisuOption':
        """Select an option by path"""
        return TiramisuOption(self._path + '.' + name,
                              index,
                              self._config_bag,
                              )


class TiramisuOptionOwner(CommonTiramisuOption):
    #FIXME optiondescription must not have Owner!
    """Manage option's owner"""

    @option_type(['symlink', 'option', 'with_index'])
    def get(self, options_bag: List[OptionBag]):
        """Get owner for a specified option"""
        option_bag = options_bag[-1]
        if len(options_bag) > 1:
            parent_option_bag = options_bag[-2]
        else:
            parent_option_bag = None
        return self._config_bag.context.get_owner(option_bag)

    @option_type(['symlink', 'option', 'with_index'])
    def isdefault(self, options_bag: List[OptionBag]):
        """Is option has defaut value"""
        option_bag = options_bag[-1]
        return self._config_bag.context.get_values().is_default_owner(option_bag)

    @option_type(['option', 'with_index'])
    def set(self,
            options_bag: List[OptionBag],
            owner: str,
            ) -> None:
        """Get owner for a specified option"""
        option_bag = options_bag[-1]
        try:
            obj_owner = getattr(owners, owner)
        except AttributeError:
            owners.addowner(owner)
            obj_owner = getattr(owners, owner)
        self._config_bag.context.get_values().set_owner(option_bag,
                                                        obj_owner,
                                                        )


class TiramisuOptionProperty(CommonTiramisuOption):
    """Manage option's property"""
    _validate_properties = False

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def get(self,
            options_bag: List[OptionBag],
            only_raises=False,
            uncalculated=False,
            ):
        """Get properties for an option"""
        option_bag = options_bag[-1]
        if not only_raises:
            return option_bag.properties
        settings = self._config_bag.context.get_settings()
        ret = settings.calc_raises_properties(option_bag,
                                              uncalculated=uncalculated,
                                              )
        return ret

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def add(self,
            options_bag: List[OptionBag],
            prop,):
        """Add new property for an option"""
        option_bag = options_bag[-1]
        if prop in FORBIDDEN_SET_PROPERTIES:
            raise ConfigError(_('cannot add this property: "{0}"').format(
                ' '.join(prop)))
        settings = self._config_bag.context.get_settings()
        props = settings.get_stored_properties(option_bag.path,
                                               option_bag.index,
                                               option_bag.option.impl_getproperties(),
                                               )
        settings.setproperties(option_bag,
                               props | {prop},
                               )

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def remove(self,
               options_bag: List[OptionBag],
               prop,
               ):
        """Remove new property for an option"""
        option_bag = options_bag[-1]
        props = option_bag.properties
        self._config_bag.context.get_settings().setproperties(option_bag,
                                                              props - {prop},
                                                              )

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def reset(self, options_bag: List[OptionBag]):
        """Reset all personalised properties"""
        option_bag = options_bag[-1]
        self._config_bag.context.get_settings().reset(option_bag)


class TiramisuOptionPermissive(CommonTiramisuOption):
    """Manage option's permissive"""

    @option_type(['option', 'optiondescription', 'symlink', 'with_or_without_index'])
    def get(self, options_bag: List[OptionBag]):
        """Get permissives value"""
        return self._config_bag.context.get_settings().getpermissives(options_bag[-1])

    @option_type(['option', 'optiondescription', 'with_or_without_index'])
    def set(self,
            options_bag: List[OptionBag],
            permissives,
            ):
        """Set permissives value"""
        option_bag = options_bag[-1]
        self._config_bag.context.get_settings().setpermissives(option_bag,
                                                               permissives=permissives,
                                                               )

    @option_type(['option', 'optiondescription', 'with_index'])
    def reset(self, options_bag: List[OptionBag]):
        """Reset all personalised permissive"""
        option_bag = options_bag[-1]
        self._config_bag.context.get_settings().reset_permissives(option_bag)


class TiramisuOptionInformation(CommonTiramisuOption):
    """Manage option's informations"""

    @option_type(['option', 'optiondescription', 'with_or_without_index', 'symlink'])
    def get(self,
            options_bag: List[OptionBag],
            name: str,
            default=undefined,
            ) -> Any:
        """Get information"""
        option_bag = options_bag[-1]
        try:
            return self._config_bag.context.get_values().get_information(option_bag,
                                                                         name,
                                                                         undefined,
                                                                         )
        except ValueError:
            return option_bag.option.impl_get_information(name, default)

    @option_type(['option', 'optiondescription'])
    def set(self,
            options_bag: List[OptionBag],
            key: str,
            value: Any) -> None:
        """Set information"""
        option_bag = options_bag[-1]
        self._config_bag.context.get_values().set_information(option_bag,
                                                              key,
                                                              value,
                                                              )

    @option_type(['option', 'optiondescription'])
    def reset(self,
              options_bag: List[OptionBag],
              key: str,
              ) -> None:
        """Remove information"""
        option_bag = options_bag[-1]
        self._config_bag.context.get_values().del_information(key,
                                                              path=option_bag.path,
                                                              )

    @option_type(['option', 'optiondescription', 'with_or_without_index', 'symlink'])
    def list(self,
             options_bag: List[OptionBag],
             ) -> list:
        """List information's keys"""
        option_bag = options_bag[-1]
        lst1 = set(option_bag.option.impl_list_information())
        lst2 = set(self._config_bag.context.get_values().list_information(option_bag.path))
        return lst1 | lst2


class TiramisuOptionValue(CommonTiramisuOption):
    """Manage option's value"""
    _validate_properties = True

    @option_type('optiondescription')
    def dict(self, options_bag: List[OptionBag]):
        """Obsolete: dict with path as key and value"""
        return self._config_bag.context.make_dict(options_bag[-1])

    @option_type(['option', 'symlink', 'with_index', 'optiondescription'])
    def get(self,
            options_bag: List[OptionBag],
            ):
        """Get value"""
        if options_bag[-1].option.impl_is_optiondescription():
            return self._config_bag.context.make_dict(options_bag[-1])
        return self._get(options_bag)

    def _get(self,
             options_bag: OptionBag,
             ):
        """Get option's value"""
        option_bag = options_bag[-1]
        if len(options_bag) > 1:
            parent_option_bag = options_bag[-2]
        else:
            parent_option_bag = None
        return self._config_bag.context.get_value(option_bag,
                                                  parent_option_bag,
                                                  )

    @option_type(['option', 'with_index'])
    def set(self,
            options_bag: List[OptionBag],
            value,
            ):
        """Change option's value"""
        option_bag = options_bag[-1]
        values = option_bag.config_bag.context.get_values()
        if not isinstance(value, Calculation) and option_bag.option.impl_is_leader() and \
                len(value) < self._config_bag.context.get_length_leadership(options_bag[-2]):
            raise LeadershipError(_('cannot reduce length of the leader "{}"'
                                    '').format(option_bag.option.impl_get_display_name()))
        return option_bag.config_bag.context.set_value(option_bag,
                                                       value,
                                                       )

    @option_type(['group', 'option', 'with_index'])
    def reset(self,
              options_bag: List[OptionBag],
              is_group: bool=False,
              ) -> None:
        """Reset value for an option"""
        option_bag = options_bag[-1]
        if is_group:
            self._config_bag.context.reset(option_bag.path,
                                           self._config_bag,
                                           )
        else:
            values = self._config_bag.context.get_values()
            if option_bag.index is not None:
                values.reset_follower(option_bag)
            else:
                values.reset(option_bag)

    @option_type(['option', 'with_index', 'symlink'])
    def default(self, options_bag: List[OptionBag]):
        """Get default value (default of option or calculated value)"""
        return self._config_bag.context.get_values().get_default_value(options_bag[-1])

    @option_type(['option', 'with_index'])
    def valid(self, options_bag: List[OptionBag]):
        """The if the option's value is valid"""
        option_bag = options_bag[-1]
        try:
            with catch_warnings(record=True) as warns:
                simplefilter("always", ValueErrorWarning)
                self._get(options_bag)
                for warn in warns:
                    if isinstance(warn.message, ValueErrorWarning):
                        return False
        except ValueError:
            return False
        return True

    @option_type(['choice', 'with_or_without_index'])
    def list(self, options_bag: List[OptionBag]):
        """All values available for a ChoiceOption"""
        option_bag = options_bag[-1]
        return option_bag.option.impl_get_values(option_bag)

    @option_type('leader')
    def pop(self,
            options_bag: List[OptionBag],
            index: int,
            ):
        """Pop a value"""
        self._config_bag.context.get_values().reset_leadership(options_bag[-1],
                                                               options_bag[-2],
                                                               index,
                                                               )

    @option_type(['leader', 'follower', 'with_or_without_index'])
    def len(self, options_bag: List[OptionBag]):
        """Length for a follower option"""
        return self._config_bag.context.get_length_leadership(options_bag[-2])


def _registers(_registers: Dict[str, type],
               prefix: str,
               ):
    for module_name in globals().keys():
        if module_name != prefix and module_name.startswith(prefix):
            module = globals()[module_name]
            func_name = module_name[len(prefix):].lower()
            _registers[func_name] = module
#__________________________________________________________________________________________________
#


class TiramisuConfig(TiramisuHelp, _TiramisuOptionWalk):
    def __init__(self,
            config_bag: ConfigBag,
            orig_config_bags: Optional[List[OptionBag]],
            ) -> None:
        self._config_bag = config_bag
        self._orig_config_bags = orig_config_bags

    def _return_config(self, config):
        if isinstance(config, KernelConfig):
            return Config(config)
        if isinstance(config, KernelMetaConfig):
            return MetaConfig(config)
        if isinstance(config, KernelMixConfig):
            return MixConfig([], config)
        if isinstance(config, KernelGroupConfig):
            return GroupConfig(config)

    def _reset_config_properties(self):
        config = self._config_bag.context
        settings = config.get_settings()
        properties = settings.get_context_properties(config.properties_cache)
        permissives = settings.get_context_permissives()
        self._config_bag.properties = properties
        self._config_bag.permissives = permissives
        if self._orig_config_bags:
            for config_bag in self._orig_config_bags:
                config_bag.properties = properties
                config_bag.permissives = permissives

    def name(self):
        """get the name"""
        return self._config_bag.context.impl_getname()


class TiramisuOption(CommonTiramisu, _TiramisuOptionOption, TiramisuConfig):
    """Manage selected option"""
    _validate_properties = False
    _registers = {}
    def __init__(self,
                 path: Optional[str]=None,
                 index: Optional[int]=None,
                 config_bag: Optional[ConfigBag]=None,
                 ) -> None:
        self._path = path
        self._index = index
        self._config_bag = config_bag
        self._tiramisu_dict = None
        if not self._registers:
            _registers(self._registers, 'TiramisuOption')

    def __getattr__(self, subfunc: str) -> Any:
        if subfunc in self._registers:
            return self._registers[subfunc](self._path,
                                            self._index,
                                            self._config_bag,
                                            )
        raise ConfigError(_(f'please specify a valid sub function ({self.__class__.__name__}.{subfunc})'))

    @option_type('optiondescription')
    def find(self,
             options_bag: List[OptionBag],
             name: str,
             value=undefined,
             type=None,
             first: bool=False):
        """Find an option by name (only for optiondescription)"""
        option_bag = options_bag[-1]
        if not first:
            ret = []
        for path in self._config_bag.context.find(option_bag=option_bag,
                                                  byname=name,
                                                  byvalue=value,
                                                  bytype=type,
                                                  ):
            t_option = TiramisuOption(path,
                                      None,  # index for a follower ?
                                      self._config_bag,
                                      )
            if first:
                return t_option
            ret.append(t_option)
        return ret

    @option_type('optiondescription')
    def group_type(self, options_bag: List[OptionBag]):
        """Get type for an optiondescription (only for optiondescription)"""
        return options_bag[-1].option.impl_get_group_type()

    @option_type('optiondescription')
    def list(self,
             options_bag: List[OptionBag],
             type='option',
             recursive=False,
             group_type=None,
             ):
        """List options inside an option description (by default list only option)"""
        return self._list(options_bag[-1],
                          type,
                          group_type,
                          recursive,
                          )

    def _load_dict(self,
                   clearable: str="all",
                   remotable: str="minimum",
                   ):
        config = self._config_bag.context
        self._tiramisu_dict = TiramisuDict(self._return_config(config),
                                           root=self._path,
                                           clearable=clearable,
                                           remotable=remotable)

    @option_type('optiondescription')
    def dict(self,
             options_bag: List[OptionBag],
             clearable: str="all",
             remotable: str="minimum",
             form: List=[],
             force: bool=False,
             ) -> Dict:
        """Convert config and option to tiramisu format"""
        if force or self._tiramisu_dict is None:
            self._load_dict(clearable, remotable)
        return self._tiramisu_dict.todict(form)

    @option_type('optiondescription')
    def updates(self,
                options_bag: List[OptionBag],
                body: List,
                ) -> Dict:
        """Updates value with tiramisu format"""
        if self._tiramisu_dict is None:  # pragma: no cover
            self._load_dict()
        return self._tiramisu_dict.set_updates(body)


class TiramisuContextInformation(TiramisuConfig):
    """Manage config informations"""
    def get(self,
            name,
            default=undefined,
            ):
        """Get an information"""
        values = self._config_bag.context.get_values()
        try:
            return values.get_information(None,
                                          name,
                                          undefined,
                                          )
        except ValueError:
            return self._config_bag.context.get_description().impl_get_information(name, default)

    def set(self,
            name,
            value,
            ):
        """Set an information"""
        self._config_bag.context.impl_set_information(self._config_bag,
                                                      name,
                                                      value,
                                                      )

    def reset(self,
              name,
              ):
        """Remove an information"""
        self._config_bag.context.impl_del_information(name)

    def list(self):
        """List information's keys"""
        lst1 = set(self._config_bag.context.get_description().impl_list_information())
        lst2 = set(self._config_bag.context.impl_list_information())
        return lst1 | lst2

    def exportation(self):
        """Export all informations"""
        return deepcopy(self._config_bag.context.get_values()._informations)

    def importation(self, informations):
        """Import informations"""
        self._config_bag.context.get_values()._informations = deepcopy(informations)


class TiramisuContextValue(TiramisuConfig):
    """Manage config value"""
    def mandatory(self):
        """Return path of options with mandatory property without any value"""
        config_bag = self._config_bag.copy()
        config_bag.properties -= {'mandatory', 'empty', 'warnings'}
        config_bag.set_permissive()
        root_option_bag = OptionBag(self._config_bag.context.get_description(),
                                    None,
                                    config_bag,
                                    )
        options = []
        for option_bag in self._config_bag.context.walk(root_option_bag,
                                                        recursive=True,
                                                        types=['mandatory'],
                                                        ):
            options.append(TiramisuOption(option_bag.path,
                                          option_bag.index,
                                          self._config_bag,
                                          ))
        return options

    # FIXME should be only for group/meta
    def set(self,
            path: str,
            value: Any,
            only_config=undefined,
            force_default=undefined,
            force_default_if_same=undefined,
            force_dont_change_value=undefined,
            ):
        """Set a value in config or children for a path"""
        kwargs = {}
        if only_config is not undefined:
            kwargs['only_config'] = only_config
        if force_default is not undefined:
            kwargs['force_default'] = force_default
        if force_default_if_same is not undefined:
            kwargs['force_default_if_same'] = force_default_if_same
        if force_dont_change_value is not undefined:
            kwargs['force_dont_change_value'] = force_dont_change_value
        option_bag = OptionBag(None,
                               None,
                               self._config_bag,
                               path=path,
                               )
        return self._config_bag.context.set_value(option_bag,
                                                  value,
                                                  **kwargs,
                                                  )

    # FIXME should be only for group/meta
    def reset(self,
              path: str,
              only_children: bool=False):
        """Reset value"""
        self._config_bag.context.reset(path,
                                       only_children,
                                       self._config_bag,
                                       )

    def dict(self):
        """Obsolete: dict with path as key and value"""
        option_bag = OptionBag(self._config_bag.context.get_description(),
                               None,
                               self._config_bag,
                               )
        return self._config_bag.context.make_dict(option_bag)

    def get(self):
        """Dict with path as key and value"""
        option_bag = OptionBag(self._config_bag.context.get_description(),
                               None,
                               self._config_bag,
                               )
        return self._config_bag.context.make_dict(option_bag)

    def exportation(self,
                    with_default_owner: bool=False,
                    ):
        """Export all values"""
        exportation = deepcopy(self._config_bag.context.get_values()._values)
        if not with_default_owner:
            del exportation[None]
        return exportation

    def importation(self, values):
        """Import values"""
        cvalues = self._config_bag.context.get_values()
        if None not in values:
            current_owner = cvalues.get_context_owner()
        cvalues._values = deepcopy(values)
        self._config_bag.context.reset_cache(None, None)
        if None not in values:
            cvalues._values[None] = {None: [None, current_owner]}


class TiramisuContextOwner(TiramisuConfig):
    """Global owner"""
    def get(self):
        """Get owner"""
        return self._config_bag.context.get_values().get_context_owner()

    def set(self, owner):
        """Set owner"""
        try:
            obj_owner = getattr(owners, owner)
        except AttributeError:
            owners.addowner(owner)
            obj_owner = getattr(owners, owner)
        values = self._config_bag.context.get_values()
        values.set_context_owner(obj_owner)


class TiramisuContextProperty(TiramisuConfig):
    """Manage config properties"""
    def read_only(self):
        """Set config to read only mode"""
        old_props = self._config_bag.properties
        settings = self._config_bag.context.get_settings()
        settings.read_only(self._config_bag)
        self._reset_config_properties()
        if 'force_store_value' not in old_props and \
                'force_store_value' in self._config_bag.properties:
            self._force_store_value()

    def read_write(self):
        """Set config to read and write mode"""
        old_props = self._config_bag.properties
        settings = self._config_bag.context.get_settings()
        settings.read_write(self._config_bag)
        or_properties = settings.rw_append - settings.ro_append - SPECIAL_PROPERTIES
        permissives = frozenset(settings.get_context_permissives() | or_properties)
        settings.set_context_permissives(permissives)
        self._reset_config_properties()
        if 'force_store_value' not in old_props and \
                'force_store_value' in self._config_bag.properties:
            self._force_store_value()

    def add(self, prop):
        """Add a config property"""
        props = set(self.get())
        if prop not in props:
            props.add(prop)
            self._set(frozenset(props))

    def remove(self, prop):
        """Remove a config property"""
        props = set(self.get())
        if prop in props:
            props.remove(prop)
            self._set(frozenset(props))

    def get(self):
        """Get all config properties"""
        return self._config_bag.properties

    def _set(self,
             props,
             ):
        """Personalise config properties"""
        if 'force_store_value' in props:
            force_store_value = 'force_store_value' not in self._config_bag.properties
        else:
            force_store_value = False
        context = self._config_bag.context
        context.get_settings().set_context_properties(props,
                                                      self._config_bag.context,
                                                      )
        self._reset_config_properties()
        if force_store_value:
            self._force_store_value()

    def reset(self):
        """Remove config properties"""
        context = self._config_bag.context
        context.get_settings().reset(self._config_bag)
        self._reset_config_properties()

    def exportation(self):
        """Export config properties"""
        return deepcopy(self._config_bag.context.get_settings()._properties)

    def importation(self, properties):
        """Import config properties"""
        if 'force_store_value' in properties.get(None, {}).get(None, []):
            force_store_value = 'force_store_value' not in self._config_bag.properties
        else:
            force_store_value = False
        self._config_bag.context.get_settings()._properties = deepcopy(properties)
        self._config_bag.context.reset_cache(None, None)
        self._reset_config_properties()
        if force_store_value:
            self._force_store_value()

    def _force_store_value(self):
        descr = self._config_bag.context.get_description()
        descr.impl_build_force_store_values(self._config_bag)

    def setdefault(self,
                   properties: Set[str],
                   type: Optional[str]=None,
                   when: Optional[str]=None) -> None:
        if not isinstance(properties, frozenset):
            raise TypeError(_('properties must be a frozenset'))
        setting = self._config_bag.context.get_settings()
        if type is None and when is None:
            setting.default_properties = properties
        else:
            if when not in ['append', 'remove']:
                raise ValueError(_('unknown when {} (must be in append or remove)').format(when))
            if type == 'read_only':
                if when == 'append':
                    setting.ro_append = properties
                else:
                    setting.ro_remove = properties
            elif type == 'read_write':
                if when == 'append':
                    setting.rw_append = properties
                else:
                    setting.rw_remove = properties
            else:
                raise ValueError(_('unknown type {}').format(type))

    def default(self,
                type: Optional[str]=None,
                when: Optional[str]=None,
                ) -> Set[str]:
        setting = self._config_bag.context.get_settings()
        if type is None and when is None:
            return setting.default_properties
        if type == 'current':
            return setting.get_context_properties(self._config_bag.context.properties_cache)
        if when not in ['append', 'remove']:
            raise ValueError(_('unknown when {} (must be in append or remove)').format(when))
        if type == 'read_only':
            if when == 'append':
                return setting.ro_append
            return setting.ro_remove
        if type == 'read_write':
            if when == 'append':
                return setting.rw_append
            return setting.rw_remove
        raise ValueError(_('unknown type {}').format(type))


class TiramisuContextPermissive(TiramisuConfig):
    """Manage config permissives"""
    def get(self):
        """Get config permissives"""
        return self._get()

    def _get(self):
        return self._config_bag.context.get_settings().get_context_permissives()

    def _set(self,
             permissives,
             ):
        """Set config permissives"""
        self._config_bag.context.get_settings().set_context_permissives(permissives)
        self._reset_config_properties()

    def exportation(self):
        """Export config permissives"""
        return deepcopy(self._config_bag.context.get_settings()._permissives)

    def importation(self, permissives):
        """Import config permissives"""
        settings = self._config_bag.context.get_settings()
        self._config_bag.context.get_settings()._permissives = deepcopy(permissives)
        self._config_bag.context.reset_cache(None,
                                             None,
                                             )
        self._reset_config_properties()

    def reset(self):
        """Remove config permissives"""
        context = self._config_bag.context
        settings = context.get_settings()
        settings.reset_permissives(self._config_bag)
        self._reset_config_properties()

    def add(self, prop):
        """Add a config permissive"""
        props = set(self._get())
        props.add(prop)
        self._set(frozenset(props))

    def remove(self, prop):
        """Remove a config permissive"""
        props = set(self._get())
        if prop in props:
            props.remove(prop)
            self._set(frozenset(props))


class TiramisuContextOption(TiramisuConfig, _TiramisuOptionWalk):
    def __init__(self,
                 *args,
                 **kwargs) -> None:
        self._tiramisu_dict = None
        super().__init__(*args, **kwargs)

    def find(self,
             name,
             value=undefined,
             type=None,
             first=False):
        """Find an or a list of options"""
        options = []
        option_bag = OptionBag(self._config_bag.context.get_description(),
                               None,
                               self._config_bag,
                               )
        for path in self._config_bag.context.find(option_bag,
                                                  byname=name,
                                                  byvalue=value,
                                                  bytype=type,
                                                  ):
            option = TiramisuOption(path,
                                    None,
                                    self._config_bag,
                                    )
            if first:
                return option
            options.append(option)
        return options

    def list(self,
             type='option',
             group_type=None,
             recursive=False,
             ):
        """List options (by default list only option)"""
        root_option_bag = OptionBag(self._config_bag.context.get_description(),
                                    None,
                                    self._config_bag,
                                    )
        return self._list(root_option_bag,
                          type,
                          group_type,
                          recursive,
                          )

    def _load_dict(self,
                   clearable="all",
                   remotable="minimum"):
        self._tiramisu_dict = TiramisuDict(self._return_config(self._config_bag.context),
                                           root=None,
                                           clearable=clearable,
                                           remotable=remotable,
                                           )

    def dict(self,
             clearable="all",
             remotable="minimum",
             form=None,
             force=False,
             ):
        """Convert config and option to tiramisu format"""
        if form is None:
            form = []
        if force or self._tiramisu_dict is None:
            self._load_dict(clearable, remotable)
        return self._tiramisu_dict.todict(form)

    def updates(self,
                body: List) -> Dict:
        """Updates value with tiramisu format"""
        if self._tiramisu_dict is None:  # pragma: no cover
            self._load_dict()
        return self._tiramisu_dict.set_updates(body)


class _TiramisuContextConfigReset():
    def reset(self):
        """Remove all datas to current config (informations, values, properties, ...)"""
        # Option's values
        context_owner = self._config_bag.context.get_values().get_context_owner()
        self._config_bag.context.get_values()._values = {None: {None: [None, context_owner]}}
        # Option's informations
        self._config_bag.context.get_values()._informations = {}
        # Option's properties
        self._config_bag.context.get_settings()._properties = {}
        # Option's permissives
        self._config_bag.context.get_settings()._permissives = {}
        # Remove cache
        self._config_bag.context.reset_cache(None, None)


class _TiramisuContextConfig(TiramisuConfig, _TiramisuContextConfigReset):
    """Actions to Config"""
    def type(self):
        """Type a Config"""
        return 'config'

    def copy(self, name=None):
        """Copy current config"""
        config = self._config_bag.context.duplicate(name=name)
        return self._return_config(config)

    def deepcopy(self, metaconfig_prefix=None, name=None):
        """Copy current config with all parents"""
        config = self._config_bag.context.duplicate(metaconfig_prefix=metaconfig_prefix,
                                                    deep=[],
                                                    name=name,
                                                    )
        return self._return_config(config)

    def parents(self):
        """Get all parents of current config"""
        ret = []
        for parent in self._config_bag.context.get_parents():
            ret.append(self._return_config(parent))
        return ret

    def path(self):
        """Get path from config (all parents name)"""
        return self._config_bag.context.get_config_path()


class _TiramisuContextGroupConfig(TiramisuConfig):
    """Actions to GroupConfig"""
    def type(self):
        """Type a Config"""
        return 'groupconfig'

    def list(self):
        """List children's config"""
        ret = []
        for child in self._config_bag.context.get_children():
            ret.append(self._return_config(child))
        return ret

    def find(self,
             name: str,
             value=undefined,
             ):
        """Find an or a list of config with finding option"""
        return GroupConfig(self._config_bag.context.find_group(byname=name,
                                                               byvalue=value,
                                                               config_bag=self._config_bag,
                                                               ))

    def __call__(self,
                 path: Optional[str]):
        """Select a child Tiramisu config"""
        spaths = path.split('.')
        config = self._config_bag.context
        for spath in spaths:
            config = config.getconfig(spath)
        if isinstance(config, KernelGroupConfig):
            return self._return_config(config)
        return self._return_config(config)


    def copy(self, name=None):
        config = self._config_bag.context.duplicate(name=name)
        return self._return_config(config)

    def deepcopy(self, name=None, metaconfig_prefix=None):
        config = self._config_bag.context.duplicate(metaconfig_prefix=metaconfig_prefix,
                                                    deep=[],
                                                    name=name,
                                                    )
        return self._return_config(config)

    def path(self):
        return self._config_bag.context.get_config_path()


class _TiramisuContextMixConfig(_TiramisuContextGroupConfig, _TiramisuContextConfigReset):
    """Actions to MixConfig"""
    def type(self):
        """Type a Config"""
        return 'mixconfig'

    def new(self, name=None, type='config'):
        """Create and add a new config"""
        config = self._config_bag.context
        new_config = config.new_config(type_=type, name=name)
        return self._return_config(new_config)

    def remove(self, name):
        """Remove config from MetaConfig"""
        config = self._config_bag.context.remove_config(name)
        return self._return_config(config)

    def add(self,
            config):
        """Add config from MetaConfig"""
        # pylint: disable=protected-access
        self._config_bag.context.add_config(config._config_bag.context)

    def parents(self):
        """Get all parents of current config"""
        ret = []
        for parent in self._config_bag.context.get_parents():
            ret.append(self._return_config(parent))
        return ret


class _TiramisuContextMetaConfig(_TiramisuContextMixConfig):
    """Actions to MetaConfig"""
    def type(self):
        """Type a Config"""
        return 'metaconfig'


class TiramisuContextCache(TiramisuConfig):
    """Manage config cache"""

    def reset(self):
        """Reset cache"""
        self._config_bag.context.reset_cache(None, None)

    def set_expiration_time(self,
                            time: int,
                            ) -> None:
        """Change expiration time value"""
        self._config_bag.expiration_time = time

    def get_expiration_time(self) -> int:
        """Get expiration time value"""
        return self._config_bag.expiration_time


class TiramisuAPI(TiramisuHelp):
    """TiramisuAPI common class
    """
    _registers = {}

    def __init__(self,
                 config_bag,
                 orig_config_bags=None) -> None:
        self._config_bag = config_bag
        self._orig_config_bags = orig_config_bags
        if not self._registers:
            _registers(self._registers, 'TiramisuContext')

    def __getattr__(self, subfunc: str) -> Any:
        if subfunc == 'option':
            config_bag = self._config_bag
            return TiramisuDispatcherOption(config_bag,
                                            self._orig_config_bags,
                                            )
        if subfunc in ['forcepermissive', 'unrestraint', 'nowarnings']:
            if self._orig_config_bags:
                msg = _('do not use unrestraint, nowarnings or forcepermissive together')
                raise ConfigError(msg)
            config_bag = self._config_bag.copy()
            if subfunc == 'unrestraint':
                config_bag.unrestraint()
            elif subfunc == 'nowarnings':
                config_bag.nowarnings()
            else:
                config_bag.set_permissive()
            return TiramisuAPI(config_bag, [self._config_bag])
        if subfunc == 'config':
            config_type = self._config_bag.context.impl_type
            if config_type == 'group':
                config = _TiramisuContextGroupConfig
            elif config_type == 'meta':
                config = _TiramisuContextMetaConfig
            elif config_type == 'mix':
                config = _TiramisuContextMixConfig
            else:
                config = _TiramisuContextConfig
            return config(self._config_bag,
                          self._orig_config_bags)
        if subfunc in self._registers:
            config_bag = self._config_bag
            # del config_bag.permissives
            return self._registers[subfunc](config_bag,
                                            self._orig_config_bags)
        raise ConfigError(_(f'please specify a valid sub function ({self.__class__.__name__}.{subfunc})'))

    def __dir__(self):
        return list(self._registers.keys()) + \
                ['unrestraint', 'forcepermissive', 'nowarnings', 'config']


class TiramisuDispatcherOption(TiramisuContextOption):
    """Select an option"""
    def __call__(self,
                 path: str,
                 index: Optional[int]=None,
                 ) -> TiramisuOption:
        """Select an option by path"""
        return TiramisuOption(path,
                              index,
                              self._config_bag,
                              )


class Config(TiramisuAPI):
    """Root config object that enables us to handle the configuration options"""
    def __init__(self,
                 descr: OptionDescription,
                 name=None,
                 display_name=None,
                 ) -> None:
        if isinstance(descr, KernelConfig):
            config = descr
        else:
            config = KernelConfig(descr,
                                  name=name,
                                  display_name=display_name,
                                  )
        settings = config.get_settings()
        properties = settings.get_context_properties(config.properties_cache)
        permissives = settings.get_context_permissives()
        config_bag = ConfigBag(config,
                               properties=properties,
                               permissives=permissives,
                               )
        super().__init__(config_bag)

    def __del__(self):
        try:
            del self._config_bag.context
            del self._config_bag
            del self._orig_config_bags
        except ConfigError:
            pass


class MetaConfig(TiramisuAPI):
    """MetaConfig object that enables us to handle the sub configuration's options
    with common root optiondescription
    """
    # pylint: disable=too-few-public-methods
    def __init__(self,
                 children: 'Config'=None,
                 name=None,
                 optiondescription: Optional[OptionDescription]=None,
                 display_name=None
                 ) -> None:
        if children is None:
            children = []
        if isinstance(children, KernelMetaConfig):
            config = children
        else:
            _children = []
            for child in children:
                if isinstance(child, TiramisuAPI):
                    _children.append(child._config_bag.context)
                else:
                    _children.append(child)

            config = KernelMetaConfig(_children,
                                      optiondescription=optiondescription,
                                      name=name,
                                      display_name=display_name,
                                      )
        settings = config.get_settings()
        properties = settings.get_context_properties(config.properties_cache)
        permissives = settings.get_context_permissives()
        config_bag = ConfigBag(config,
                               properties=properties,
                               permissives=permissives)
        super().__init__(config_bag)


class MixConfig(TiramisuAPI):
    """MixConfig object that enables us to handle the sub configuration's options
    with differents root optiondescription
    """
    # pylint: disable=too-few-public-methods
    def __init__(self,
                 optiondescription: OptionDescription,
                 children: List[Config],
                 name: Callable=None,
                 display_name=None
                 ) -> None:
        if isinstance(children, KernelMixConfig):
            config = children
        else:
            _children = []
            for child in children:
                if isinstance(child, TiramisuAPI):
                    _children.append(child._config_bag.context)
                else:
                    _children.append(child)

            config = KernelMixConfig(optiondescription,
                                     _children,
                                     name=name,
                                     display_name=display_name,
                                     )
        settings = config.get_settings()
        properties = settings.get_context_properties(config.properties_cache)
        permissives = settings.get_context_permissives()
        config_bag = ConfigBag(config,
                               properties=properties,
                               permissives=permissives,
                               )
        super().__init__(config_bag)


class GroupConfig(TiramisuAPI):
    """GroupConfig that enables us to access the sub configuration's options"""
    # pylint: disable=too-few-public-methods
    def __init__(self,
                 children,
                 name=None,
                 ) -> None:
        if isinstance(children, KernelGroupConfig):
            config = children
        else:
            _children = []
            for child in children:
                if isinstance(child, TiramisuAPI):
                    _children.append(child._config_bag.context)
                else:
                    _children.append(child)

            config = KernelGroupConfig(_children, name=name)
        config_bag = ConfigBag(config,
                               properties=None,
                               permissives=None)
        super().__init__(config_bag)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""DynOptionDescription
"""
import re
import weakref
from typing import List, Any, Optional, Tuple
from itertools import chain
from ..autolib import ParamOption


from ..i18n import _
from .optiondescription import OptionDescription
from .syndynoptiondescription import SynDynLeadership
from .baseoption import BaseOption
from ..setting import OptionBag, ConfigBag, undefined
from ..error import ConfigError
from ..autolib import Calculation


NAME_REGEXP = re.compile(r'^[a-zA-Z\d\-_]*$')


class DynOptionDescription(OptionDescription):
    """dyn option description
    """
    __slots__ = ('_suffixes',
                 '_subdyns',
                 )

    def __init__(self,
                 name: str,
                 doc: str,
                 children: List[BaseOption],
                 suffixes: Calculation,
                 properties=None,
                 ) -> None:
        # pylint: disable=too-many-arguments
        super().__init__(name,
                         doc,
                         children,
                         properties,
                         )
        # check children + set relation to this dynoptiondescription
        wself = weakref.ref(self)
        for child in children:
            child._setsubdyn(wself)
        # add suffixes
        if __debug__ and not isinstance(suffixes, Calculation):
            raise ConfigError(_('suffixes in dynoptiondescription has to be a calculation'))
        for param in chain(suffixes.params.args, suffixes.params.kwargs.values()):
            if isinstance(param, ParamOption):
                param.option._add_dependency(self,
                                             is_suffix=True,
                                             )
        self._suffixes = suffixes

    def convert_suffix_to_path(self,
                               suffix: Any,
                               ) -> str:
        """convert suffix to use it to a path
        """
        if suffix is None:
            return None
        if not isinstance(suffix, str):
            suffix = str(suffix)
        if '.' in suffix:
            suffix = suffix.replace('.', '_')
        return suffix

    def get_suffixes(self,
                     config_bag: ConfigBag,
                     dynoption=None,
                     ) -> List[str]:
        """get dynamic suffixes
        """
        if dynoption:
            self_opt = dynoption
        else:
            self_opt = self
        option_bag = OptionBag(self_opt,
                               None,
                               config_bag,
                               properties=None,
                               )
        values = self._suffixes.execute(option_bag)
        if values is None:
            values = []
        values_ = []
        if __debug__:
            if not isinstance(values, list):
                raise ValueError(_('DynOptionDescription suffixes for '
                                   f'option "{self.impl_get_display_name()}", is not '
                                   f'a list ({values})'))
        for val in values:
            cval = self.convert_suffix_to_path(val)
            if not isinstance(cval, str) or re.match(NAME_REGEXP, cval) is None:
                if __debug__ and cval is not None:
                    raise ValueError(_('invalid suffix "{}" for option "{}"'
                                       '').format(cval,
                                                  self.impl_get_display_name()))
            else:
                values_.append(val)
        if __debug__ and len(values_) > len(set(values_)):
            raise ValueError(_(f'DynOptionDescription "{self._name}" suffixes return a list with '
                               f'same values "{values_}"'''))
        return values_

    def impl_is_dynoptiondescription(self) -> bool:
        return True

    def option_is_self(self,
                       option,
                       ) -> bool:
        return option == self or \
                (option.impl_is_sub_dyn_optiondescription() and option.opt == self)

    def split_path(self,
                   dynoption,
                   option,
                   ) -> Tuple[str, str]:
        """self.impl_getpath() is something like root.xxx.dynoption_path
        option.impl_getpath() is something like root.xxx.dynoption_path.sub.path
        must return ('root.xxx.', '.sub')
        """
        if dynoption is None:
            self_path = self.impl_getpath()
        else:
            self_path = dynoption.impl_getpath()
        root_path = self_path.rsplit('.', 1)[0] if '.' in self_path else None
        #
        if self.option_is_self(option):
            sub_path = ''
        else:
            option_path = option.impl_getpath()
            if root_path:
                if isinstance(option, SynDynLeadership):
                    count_root_path = option_path.count('.') - root_path.count('.')
                    root_path = option_path.rsplit('.', count_root_path)[0]
                root_path += '.'
            self_number_child = self_path.count('.') + 1
            option_sub_path = option_path.split('.', self_number_child)[-1]
            sub_path = '.' + option_sub_path.rsplit('.', 1)[0] if '.' in option_sub_path else ''
        return root_path, sub_path

    def get_sub_children(self,
                         option,
                         config_bag,
                         *,
                         index=None,
                         properties=undefined,
                         dynoption=None,
                         ):
        root_path, sub_path = self.split_path(dynoption,
                                              option,
                                              )
        for suffix in self.get_suffixes(config_bag,
                                        dynoption=dynoption,
                                        ):
            if self.option_is_self(option):
                parent_path = root_path
            elif root_path:
                parent_path = root_path + self.impl_getname(suffix) + sub_path
            else:
                parent_path = self.impl_getname(suffix) + sub_path
            yield OptionBag(option.to_dynoption(parent_path,
                                                suffix,
                                                self,
                                                ),
                            index,
                            config_bag,
                            properties=properties,
                            ori_option=option
                            )

    def impl_getname(self, suffix=None) -> str:
        """get name
        """
        name = super().impl_getname()
        if suffix is None:
            return name
        path_suffix = self.convert_suffix_to_path(suffix)
        return name + path_suffix



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""options handler global entry point
"""
import weakref
from copy import copy, deepcopy
from typing import Optional, List, Any, Union

from .error import PropertiesOptionError, ConfigError, ConflictError, \
                   LeadershipError
from .option import DynOptionDescription, Leadership, Option
from .option.syndynoptiondescription import SubDynOptionDescription
from .setting import OptionBag, ConfigBag, Settings, undefined, groups
from .value import Values, owners
from .i18n import _
from .cacheobj import Cache
from .autolib import Calculation


class _SubConfig:
    """Sub configuration management entry.
    Tree if OptionDescription's responsability. SubConfig are generated
    on-demand. A Config is also a SubConfig.
    Root Config is call context below
    """
    __slots__ = ('_impl_context',
                 '_impl_descr',
                 '_impl_path',
                 )

    def __init__(self,
                 descr,
                 context,
                 subpath=None,
                 ):
        """ Configuration option management class

        :param descr: describes the configuration schema
        :type descr: an instance of ``option.OptionDescription``
        :param context: the current root config
        :type context: `Config`
        :type subpath: `str` with the path name
        """
        # main option description
        self._impl_descr = descr
        self._impl_context = context
        self._impl_path = subpath

    def get_length_leadership(self,
                              option_bag,
                              ):
        """Get the length of leader option (useful to know follower's length)
        """
        cconfig_bag = option_bag.config_bag.copy()
        cconfig_bag.remove_validation()
        option_bag = OptionBag(option_bag.option.get_leader(),
                               None,
                               cconfig_bag,
                               )
        return len(self.get_value(option_bag))

    def get_description(self):
        """get root description
        """
        assert self._impl_descr is not None, _('there is no option description for this config'
                                               ' (may be GroupConfig)')
        return self._impl_descr

    def get_settings(self):
        """get settings object
        """
        return self._impl_settings  # pylint: disable=no-member

    def get_values(self):
        """get values object
        """
        return self._impl_values  # pylint: disable=no-member

    # =============================================================================
    # CACHE
    def reset_cache(self,
                    option_bag,
                    resetted_opts=None,
                    ):
        """reset all settings in cache
        """
        if resetted_opts is None:
            resetted_opts = []

        if option_bag is not None:
            if 'cache' not in option_bag.config_bag.properties:
                return
            option_bag.config_bag.properties = option_bag.config_bag.properties - {'cache'}
            self.reset_one_option_cache(resetted_opts,
                                        option_bag,
                                        )
            option_bag.config_bag.properties = option_bag.config_bag.properties | {'cache'}
        else:
            self._impl_values_cache.reset_all_cache()  # pylint: disable=no-member
            self.properties_cache.reset_all_cache()  # pylint: disable=no-member

    def get_values_cache(self):
        """get cache for values
        """
        return self._impl_values_cache  # pylint: disable=no-member

    def reset_one_option_cache(self,
                               resetted_opts,
                               option_bag,
                               ):
        """reset cache for one option
        """
        if option_bag.path in resetted_opts:
            return
        resetted_opts.append(option_bag.path)
        for woption in option_bag.option.get_dependencies(option_bag.option):
            option = woption()
            if woption in option_bag.option._get_suffixes_dependencies() and \
                    option_bag.option.issubdyn() and \
                    option.impl_is_dynoptiondescription():
                paths = [subdyn().impl_getpath() for subdyn in option.get_sub_dyns()]
                for weak_subdyn in option_bag.option.get_sub_dyns():
                    subdyn = weak_subdyn()
                    if subdyn.impl_getpath() in paths:
                        root_path = option_bag.option.impl_getpath()
                        if '.' in root_path:
                            root_path = root_path.rsplit('.', 1)[0]
                            nb_elt = root_path.count('.') + 1
                        else:
                            root_path = ''
                            nb_elt = 1
                        config_bag = option_bag.config_bag
                        root_option_bag = OptionBag(config_bag.context.get_description(),
                                                    None,
                                                    config_bag,
                                                    )
                        full_path = root_path + '.' + option.impl_getpath().split('.', nb_elt)[-1]
                        try:
                            options_bag = config_bag.context.get_sub_option_bag(root_option_bag,
                                                                                full_path,
                                                                                #FIXME index?
                                                                                None,
                                                                                validate_properties=False,
                                                                                properties=None,
                                                                                allow_dynoption=True,
                                                                                )
                        except AttributeError as err:
                            raise ConfigError(_(f'option "{option.impl_get_display_name()}" is not in a dynoptiondescription: {err}'))
                        self._reset_cache_dyn_optiondescription(options_bag[-1],
                                                                resetted_opts,
                                                                )
                        break
            else:
                soption_bag = OptionBag(option,
                                        option_bag.index,
                                        option_bag.config_bag,
                                        properties=None,
                                        )
                if option.impl_is_dynoptiondescription():
                    self._reset_cache_dyn_optiondescription(soption_bag,
                                                            resetted_opts,
                                                            )
                elif option.issubdyn():
                    # it's an option in dynoptiondescription, remove cache for all generated option
                    config_bag = option_bag.config_bag
                    options = [soption_bag]
                    for wdynopt in reversed(option.get_sub_dyns()):
                        dynopt = wdynopt()
                        sub_options = []
                        for sub_option in options:
                            sub_options.extend(dynopt.get_sub_children(sub_option.option,
                                                                       config_bag,
                                                                       properties=None,
                                                                       ))
                        options = sub_options
                    for doption_bag in options:
                        self.reset_one_option_cache(resetted_opts,
                                                    doption_bag,
                                                    )
                else:
                    self.reset_one_option_cache(resetted_opts,
                                                soption_bag,
                                                )
            del option
        option_bag.option.reset_cache(option_bag.path,
                                      option_bag.config_bag,
                                      resetted_opts,
                                      )

    def _reset_cache_dyn_optiondescription(self,
                                           option_bag,
                                           resetted_opts,
                                           ):
        # reset cache for all chidren
        for doption_bag in option_bag.option.get_sub_children(option_bag.option,
                                                              option_bag.config_bag,
                                                              index=option_bag.index,
                                                              properties=None,
                                                              ):
            for coption in doption_bag.option.get_children_recursively(None,
                                                                       None,
                                                                       option_bag.config_bag,
                                                                       ):
                coption_bag = self.get_sub_option_bag(doption_bag,  # pylint: disable=no-member
                                                      coption.impl_getpath(),
                                                      None,
                                                      False,
                                                      )[-1]
                self.reset_one_option_cache(resetted_opts,
                                            coption_bag,
                                            )
        self._reset_cache_dyn_option(option_bag,
                                     resetted_opts,
                                     )

    def _reset_cache_dyn_option(self,
                                option_bag,
                                resetted_opts,
                                ):
        option = option_bag.option
        if isinstance(option, (DynOptionDescription, SubDynOptionDescription)):
            dynoption = option
        else:
            dynoption = option.getsubdyn()
        for doption_bag in dynoption.get_sub_children(option,
                                                      option_bag.config_bag,
                                                      index=option_bag.index,
                                                      properties=None
                                                      ):
            self.reset_one_option_cache(resetted_opts,
                                        doption_bag,
                                        )

    # =============================================================================
    # WALK
    def find(self,
             option_bag,
             bytype,
             byname,
             byvalue,
             raise_if_not_found=True,
             only_path=undefined,
             only_option=undefined,
             with_option=False,
             ):
        """
        convenience method for finding an option that lives only in the subtree

        :param first: return only one option if True, a list otherwise
        :return: find list or an exception if nothing has been found
        """
        # pylint: disable=too-many-arguments,too-many-locals
        def _filter_by_value(soption_bag):
            value = self.get_value(soption_bag)
            if isinstance(value, list):
                return byvalue in value
            return value == byvalue

        found = False
        if only_path is not undefined:
            def _fake_iter():
                yield only_option
            options = _fake_iter()
        else:
            options = option_bag.option.get_children_recursively(bytype,
                                                                 byname,
                                                                 option_bag.config_bag,
                                                                 )
        for option in options:
            path = option.impl_getpath()
            soption_bag = OptionBag(option,
                                    None,
                                    option_bag.config_bag,
                                    )
            if byvalue is not undefined and not _filter_by_value(soption_bag):
                continue
            if option_bag.config_bag.properties:
                #remove option with propertyerror, ...
                try:
                    self.get_sub_option_bag(option_bag,  # pylint: disable=no-member
                                            path,
                                            None,
                                            True,
                                            )[-1]
                except PropertiesOptionError:
                    continue
            found = True
            if not with_option:
                yield path
            else:
                yield path, option
        self._find_return_results(found,
                                  raise_if_not_found,
                                  )

    def _find_return_results(self,
                             found,
                             raise_if_not_found):
        if not found and raise_if_not_found:
            raise AttributeError(_("no option found in config"
                                   " with these criteria"))

    def _walk_valid_value(self,
                          option_bag,
                          types,
                          value=undefined,
                          ):
        if value is undefined:
            value = self.get_value(option_bag,
                                   need_help=False,
                                   )
        l_option_bag = option_bag.copy()
        l_option_bag.config_bag = option_bag.config_bag.copy()
        if 'mandatory' in types:
            l_option_bag.config_bag.properties |= {'mandatory', 'empty'}
        self.get_settings().validate_mandatory(value,
                                               l_option_bag,
                                               )
        return value

    def walk(self,
             option_bag: OptionBag,
             types: List[str]=('option',),
             group_type=None,
             recursive: bool=True,
             walked: bool=False,
             flatten_leadership: bool=False,
             ):
        """walk to tree
        """
        # pylint: disable=too-many-branches,too-many-locals,too-many-arguments,
        if option_bag.option.impl_is_optiondescription():
            # do not return root option
            if walked:
                if 'optiondescription' in types and (group_type is None or
                        option_bag.option.impl_get_group_type() == group_type):
                    yield option_bag
                if not recursive:
                    # it's not recursive, so stop to walk
                    return
            if not option_bag.option.impl_is_leadership() or flatten_leadership:
                for opt in option_bag.option.get_children(option_bag.config_bag):
                    try:
                        yield from self.walk(self.get_sub_option_bag(option_bag,  # pylint: disable=no-member
                                                                     opt.impl_getpath(),
                                                                     None,
                                                                     True,
                                                                     follower_not_apply_requires=flatten_leadership,
                                                                     )[-1],
                                             types=types,
                                             recursive=recursive,
                                             group_type=group_type,
                                             walked=True,
                                             )
                    except PropertiesOptionError as err:
                        if err.proptype in (['mandatory'], ['empty']):
                            raise err
            elif 'option' in types or 'mandatory' in types:
                # it's a leadership so walk to leader and followers
                # followers has specific length
                leader, *followers = option_bag.option.get_children(option_bag.config_bag)
                leader_option_bag = self.get_sub_option_bag(option_bag,  # pylint: disable=no-member
                                                            leader.impl_getpath(),
                                                            None,
                                                            True,
                                                            )[-1]
                followers_dict = {'leader': leader_option_bag}
                values = self.get_value(leader_option_bag,
                                        need_help=False,
                                        )
                ls_length = len(values)
                try:
                    self._walk_valid_value(leader_option_bag,
                                           types,
                                           value=values,
                                           )
                except PropertiesOptionError as err:
                    if err.proptype in (['mandatory'], ['empty']):
                        yield leader_option_bag
                for idx in range(ls_length):
                    followers_dict[idx] = []
                    for follower in followers:
                        follower_path = follower.impl_getpath()
                        try:
                            options_bag = self.get_sub_option_bag(option_bag,  # pylint: disable=no-member
                                                                  follower_path,
                                                                  idx,
                                                                  True,
                                                                  leadership_length=ls_length,
                                                                  )
                            for f_follower_bag in self.walk(options_bag[-1],
                                                            types=types,
                                                            recursive=recursive,
                                                            group_type=group_type,
                                                            walked=True,
                                                            ):
                                if 'mandatory' in types:
                                    yield f_follower_bag
                                followers_dict[idx].append(f_follower_bag)
                        except PropertiesOptionError as err:
                            continue
                if 'option' in types:
                    yield followers_dict
        else:
            if 'mandatory' in types and not option_bag.option.impl_is_symlinkoption():
                try:
                    self._walk_valid_value(option_bag,
                                           types,
                                           )
                except PropertiesOptionError as err:
                    if err.proptype in (['mandatory'], ['empty']):
                        yield option_bag
            if 'option' in types:
                yield option_bag

    def make_dict(self, option_bag):
        """exports the whole config into a `dict`
        :returns: dict of Option's name (or path) and values
        """
        ret = {}
        for data in self.walk(option_bag):
            if isinstance(data, OptionBag):
                option_bag = data
                ret[option_bag.path] = self.get_value(option_bag,
                                                      need_help=False,
                                                      )
            else:
                leader_ret = []
                leader_path = data['leader'].path
                for idx, value in enumerate(self.get_value(data['leader'],
                                                           need_help=False,
                                                           )):
                    leader_dict = {leader_path: value}
                    for follower in data.get(idx, []):
                        leader_dict[follower.path] = self.get_value(follower,
                                                                    need_help=False,
                                                                    )
                    leader_ret.append(leader_dict)
                ret[leader_path] = leader_ret

        return ret

    # =============================================================================
    # Manage value
    def set_value(self,
                  option_bag: OptionBag,
                  value: Any,
                  ) -> Any:
        """set value
        """
        self.get_settings().validate_properties(option_bag)
        return self.get_values().set_value(option_bag,
                                           value
                                           )

    def get_value(self,
                  option_bag,
                  parent_option_bag=None,
                  need_help=True,
                  ):
        """
        :return: option's value if name is an option name, OptionDescription
                 otherwise
        """
        option_bag = self._get(option_bag,
                               need_help,
                               )
        if isinstance(option_bag, list):
            value = []
            for opt_bag in option_bag:
                value.append(self.get_value(opt_bag,
                                            need_help=need_help,
                                            ))
        else:
            value = self.get_values().get_cached_value(option_bag)
            if parent_option_bag and option_bag.option.impl_is_follower() :
                length = self.get_length_leadership(parent_option_bag)
                follower_len = self.get_values().get_max_length(option_bag.path)
                if follower_len > length:
                    option_name = option_bag.option.impl_get_display_name()
                    raise LeadershipError(_(f'the follower option "{option_name}" '
                                            f'has greater length ({follower_len}) than the leader '
                                            f'length ({length})'))
            self.get_settings().validate_mandatory(value,
                                                   option_bag,
                                                   )
        return value

    def _get(self,
             option_bag: OptionBag,
             need_help: bool,
             ) -> OptionBag:
        # pylint: disable=too-many-locals
        option = option_bag.option
        if option.impl_is_symlinkoption():
            suboption = option.impl_getopt()
            if suboption.issubdyn():
                dynopt = suboption.getsubdyn()
                return list(dynopt.get_sub_children(suboption,
                                                    option_bag.config_bag,
                                                    index=option_bag.index,
                                                    ))
            if suboption.impl_is_follower():
                options_bag = self.get_sub_option_bag(option_bag.config_bag,  # pylint: disable=no-member
                                                      suboption.impl_getpath(),
                                                      None,
                                                      True,
                                                      )
                leadership_length = self.get_length_leadership(options_bag[-2])
                ret = []
                for idx in range(leadership_length):
                    f_option_bag = OptionBag(suboption,
                                             idx,
                                             option_bag.config_bag,
                                             )
                    ret.append(f_option_bag)
                return ret

            soption_bag = OptionBag(suboption,
                                    option_bag.index,
                                    option_bag.config_bag,
                                    ori_option=option,
                                    )
            return self._get(soption_bag,
                             need_help,
                             )
        return option_bag

    def get_owner(self, option_bag: OptionBag):
        """get owner
        """
        options_bag = self._get(option_bag,
                                need_help=True,
                                )
        if isinstance(options_bag, list):
            for opt_bag in options_bag:
                owner = self.get_owner(opt_bag)
                if owner != owners.default:
                    break
            else:
                owner = owners.default
        else:
            owner = self.get_values().getowner(options_bag)
        return owner


class _CommonConfig(_SubConfig):
    "abstract base class for the Config, KernelGroupConfig and the KernelMetaConfig"
    __slots__ = ('_impl_values',
                 '_impl_values_cache',
                 '_impl_settings',
                 'properties_cache',
                 '_impl_permissives_cache',
                 'parents',
                 'impl_type',
                 )

    def _impl_build_all_caches(self, descr):
        if not descr.impl_already_build_caches():
            descr._group_type = groups.root  # pylint: disable=protected-access
            descr._build_cache(display_name=self._display_name)  # pylint: disable=no-member,protected-access
        if not hasattr(descr, '_cache_force_store_values'):
            raise ConfigError(_('option description seems to be part of an other '
                                'config'))

    def get_parents(self):
        """get parents
        """
        for parent in self.parents:  # pylint: disable=no-member
            yield parent()

    # information
    def impl_set_information(self,
                             config_bag,
                             key,
                             value,
                             ):
        """updates the information's attribute

        :param key: information's key (ex: "help", "doc"
        :param value: information's value (ex: "the help string")
        """
        self._impl_values.set_information(None,  # pylint: disable=no-member
                                          key,
                                          value,
                                          )
        for option in self.get_description()._cache_dependencies_information.get(key, []):  # pylint: disable=protected-access
            option_bag = OptionBag(option,
                                   None,
                                   config_bag,
                                   properties=None,
                                   )
            self.reset_cache(option_bag)

    def impl_get_information(self,
                             key,
                             default,
                             ):
        """retrieves one information's item

        :param key: the item string (ex: "help")
        """
        return self._impl_values.get_information(None,  # pylint: disable=no-member
                                                 key,
                                                 default,
                                                 )

    def impl_del_information(self,
                             key,
                             raises=True,
                             ):
        """delete an information
        """
        self._impl_values.del_information(key,  # pylint: disable=no-member
                                          raises,
                                          )

    def impl_list_information(self):
        """list information keys for context
        """
        return self._impl_values.list_information()  # pylint: disable=no-member

    def gen_fake_values(self) -> 'KernelConfig':
        """generate a fake values to improve validation when assign a new value
        """
        export = deepcopy(self.get_values()._values)  # pylint: disable=protected-access
        fake_config = KernelConfig(self._impl_descr,
                                   force_values=export,
                                   force_settings=self.get_settings(),
                                   name=self._impl_name,  # pylint: disable=no-member
                                   )
        fake_config.parents = self.parents  # pylint: disable=no-member
        return fake_config

    def duplicate(self,
                  force_values=None,
                  force_settings=None,
                  metaconfig_prefix=None,
                  child=None,
                  deep=None,
                  name=None,
                  ):
        """duplication config
        """
        # pylint: disable=too-many-arguments
        if name is None:
            name = self._impl_name  # pylint: disable=no-member
        if isinstance(self, KernelConfig):
            duplicated_config = KernelConfig(self._impl_descr,
                                             _duplicate=True,
                                             force_values=force_values,
                                             force_settings=force_settings,
                                             name=name,
                                             )
        else:
            duplicated_config = KernelMetaConfig([],
                                                 _duplicate=True,
                                                 optiondescription=self._impl_descr,
                                                 name=name,
                                                 )
        duplicated_values = duplicated_config.get_values()
        duplicated_settings = duplicated_config.get_settings()
        duplicated_values._values = deepcopy(self.get_values()._values)  # pylint: disable=protected-access
        duplicated_values._informations = deepcopy(self.get_values()._informations)  # pylint: disable=protected-access
        duplicated_settings._properties = deepcopy(self.get_settings()._properties)  # pylint: disable=protected-access
        duplicated_settings._permissives = deepcopy(self.get_settings()._permissives)  # pylint: disable=protected-access
        duplicated_settings.ro_append = self.get_settings().ro_append
        duplicated_settings.rw_append = self.get_settings().rw_append
        duplicated_settings.ro_remove = self.get_settings().ro_remove
        duplicated_settings.rw_remove = self.get_settings().rw_remove
        duplicated_settings.default_properties = self.get_settings().default_properties
        duplicated_config.reset_cache(None, None)
        if child is not None:
            duplicated_config._impl_children.append(child)  # pylint: disable=protected-access
            child.parents.append(weakref.ref(duplicated_config))
        if self.parents:  # pylint: disable=no-member
            if deep is not None:
                for parent in self.parents:  # pylint: disable=no-member
                    wparent = parent()
                    if wparent not in deep:
                        deep.append(wparent)
                        subname = wparent.impl_getname()
                        if metaconfig_prefix:
                            subname = metaconfig_prefix + subname
                        duplicated_config = wparent.duplicate(deep=deep,
                                                              metaconfig_prefix=metaconfig_prefix,
                                                              child=duplicated_config,
                                                              name=subname,
                                                              )
            else:
                duplicated_config.parents = self.parents  # pylint: disable=no-member
                for parent in self.parents:  # pylint: disable=no-member
                    parent()._impl_children.append(duplicated_config)  # pylint: disable=protected-access
        return duplicated_config

    def get_config_path(self):
        """get config path
        """
        path = self.impl_getname()
        for parent in self.parents:  # pylint: disable=no-member
            wparent = parent()
            if wparent is None:  # pragma: no cover
                raise ConfigError(_(f'parent of {self._impl_name} not already exists'))  # pylint: disable=no-member
            path = parent().get_config_path() + '.' + path
        return path

    def get_sub_option_bag(self,
                           bag: Union[OptionBag, ConfigBag],
                           path: str,
                           index: Optional[int],
                           validate_properties: bool,
                           leadership_length: int=None,
                           *,
                           properties=undefined,
                           follower_not_apply_requires: bool=False,
                           allow_dynoption: bool=False,
                           ) -> List[OptionBag]:
        """Get the suboption for path and the name of the option
        :returns: option_bag
        """
        # pylint: disable=too-many-branches,too-many-locals,too-many-arguments
        if isinstance(bag, ConfigBag):
            option_bag = OptionBag(self.get_description(),
                                   None,
                                   bag,
                                   )
        else:
            option_bag = bag
            if option_bag.option != option_bag.config_bag.context.get_description():
                path = path[len(option_bag.path) + 1:]
        split_path = path.split('.')
        last_idx = len(split_path) - 1
        suboption = option_bag.option
        options_bag = []
        sub_option_bag = option_bag
        for idx, step in enumerate(split_path):
            if not suboption.impl_is_optiondescription():
                raise TypeError(f'{suboption.impl_getpath()} is not an optiondescription')

            if not idx and option_bag.option == self.get_description():
                subpath = None
            else:
                subpath = suboption.impl_getpath()
            option = suboption.get_child(step,
                                         option_bag.config_bag,
                                         subpath,
                                         allow_dynoption=allow_dynoption,
                                         )
            if idx == last_idx:
                option_index = index
                apply_requires = not follower_not_apply_requires or \
                        option.impl_is_optiondescription() or \
                        not option.impl_is_follower()
                if option_index is not None:
                    if option.impl_is_optiondescription() or \
                            option.impl_is_symlinkoption() or \
                            not option.impl_is_follower():
                        raise ConfigError('index must be set only with a follower option')
                    if leadership_length is not None:
                        length = leadership_length
                    else:
                        length = self.get_length_leadership(sub_option_bag)
                    if index >= length:
                        raise LeadershipError(_(f'index "{index}" is greater than the leadership '
                                                f'length "{length}" for option '
                                                f'"{option.impl_get_display_name()}"'))
                option_properties = properties
            else:
                option_index = None
                apply_requires = True
                option_properties = undefined
            sub_option_bag = OptionBag(option,
                                       option_index,
                                       option_bag.config_bag,
                                       properties=option_properties,
                                       apply_requires=apply_requires,
                                       )
            if validate_properties:
                self.get_settings().validate_properties(sub_option_bag)
            suboption = option
            options_bag.append(sub_option_bag)
        return options_bag

    def impl_getname(self):
        """get config name
        """
        return self._impl_name  # pylint: disable=no-member


# ____________________________________________________________
class KernelConfig(_CommonConfig):
    """main configuration management entry
    """
    # pylint: disable=too-many-instance-attributes
    __slots__ = ('__weakref__',
                 '_impl_name',
                 '_display_name',
                 '_impl_symlink',
                 '_storage',
                 )
    impl_type = 'config'

    def __init__(self,
                 descr,
                 force_values=None,
                 force_settings=None,
                 name=None,
                 display_name=None,
                 _duplicate=False,
                 ):
        """ Configuration option management class

        :param descr: describes the configuration schema
        :type descr: an instance of ``option.OptionDescription``
        :param context: the current root config
        :type context: `Config`
        """
        # pylint: disable=too-many-arguments,too-many-arguments
        self._display_name = display_name
        self.parents = []
        self._impl_symlink = []
        self._impl_name = name
        if isinstance(descr, Leadership):
            raise ConfigError(_('cannot set leadership object has root optiondescription'))
        if isinstance(descr, DynOptionDescription):
            msg = _('cannot set dynoptiondescription object has root optiondescription')
            raise ConfigError(msg)
        if force_settings is not None and force_values is not None:
            self._impl_settings = force_settings
            self._impl_permissives_cache = Cache()
            self.properties_cache = Cache()
            self._impl_values = Values(force_values)
            self._impl_values_cache = Cache()
        else:
            self._impl_settings = Settings()
            self._impl_permissives_cache = Cache()
            self.properties_cache = Cache()
            self._impl_values = Values()
            self._impl_values_cache = Cache()
        self._impl_context = weakref.ref(self)
        if None in [force_settings, force_values]:
            self._impl_build_all_caches(descr)
        super().__init__(descr,
                         self._impl_context,
                         None,
                         )


class KernelGroupConfig(_CommonConfig):
    """Group a config with same optiondescription tree
    """
    __slots__ = ('__weakref__',
                 '_impl_children',
                 '_impl_name',
                 '_display_name',
                 )
    impl_type = 'group'

    def __init__(self,
                 children,
                 display_name=None,
                 name=None,
                 _descr=None,
                 ):
        # pylint: disable=super-init-not-called
        names = []
        for child in children:
            name_ = child._impl_name
            names.append(name_)
        if len(names) != len(set(names)):
            while range(1, len(names) + 1):
                name = names.pop(0)
                if name in names:
                    raise ConflictError(_('config name must be uniq in '
                                          'groupconfig for "{0}"').format(name))

        self._impl_children = children
        self.parents = []
        self._display_name = display_name
        if name:
            self._impl_name = name
        self._impl_context = weakref.ref(self)
        self._impl_descr = _descr
        self._impl_path = None

    def get_children(self):
        """get all children
        """
        return self._impl_children

    def reset_cache(self,
                    option_bag,
                    resetted_opts=None,
                    ):
        if resetted_opts is None:
            resetted_opts = []
        if isinstance(self, KernelMixConfig):
            super().reset_cache(option_bag,
                                resetted_opts=copy(resetted_opts),
                                )
        for child in self._impl_children:
            if option_bag is not None:
                coption_bag = option_bag.copy()
                cconfig_bag = coption_bag.config_bag.copy()
                cconfig_bag.context = child
                coption_bag.config_bag = cconfig_bag
            else:
                coption_bag = None
            child.reset_cache(coption_bag,
                              resetted_opts=copy(resetted_opts),
                              )

    def set_value(self,
                  option_bag,
                  value,
                  only_config=False,
                  ):
        """Setattr not in current KernelGroupConfig, but in each children
        """
        ret = []
        for child in self._impl_children:
            cconfig_bag = option_bag.config_bag.copy()
            cconfig_bag.context = child
            if isinstance(child, KernelGroupConfig):
                ret.extend(child.set_value(option_bag,
                                           value,
                                           only_config=only_config,
                                           ))
            else:
                settings = child.get_settings()
                properties = settings.get_context_properties(child.properties_cache)
                permissives = settings.get_context_permissives()
                cconfig_bag.properties = properties
                cconfig_bag.permissives = permissives
                try:
                    coption_bag = child.get_sub_option_bag(cconfig_bag,
                                                           option_bag.path,
                                                           option_bag.index,
                                                           False,
                                                           )[-1]
                    child.set_value(coption_bag,
                                    value,
                                    )
                except PropertiesOptionError as err:
                    # pylint: disable=protected-access
                    ret.append(PropertiesOptionError(err._option_bag,
                                                     err.proptype,
                                                     err._settings,
                                                     err._opt_type,
                                                     err._name,
                                                     err._orig_opt))
                except (ValueError, LeadershipError, AttributeError) as err:
                    ret.append(err)
        return ret


    def find_group(self,
                   config_bag,
                   byname=None,
                   bypath=undefined,
                   byoption=undefined,
                   byvalue=undefined,
                   raise_if_not_found=True,
                   _sub=False,
                   ):
        """Find first not in current KernelGroupConfig, but in each children
        """
        # pylint: disable=too-many-arguments
        # if KernelMetaConfig, all children have same OptionDescription in
        # context so search only one time the option for all children
        if bypath is undefined and byname is not None and \
                 self.impl_type == 'meta':
            root_option_bag = OptionBag(self.get_description(),
                                        None,
                                        config_bag,
                                        )
            next(self.find(root_option_bag,
                           bytype=None,
                           byname=byname,
                           byvalue=undefined,
                           raise_if_not_found=raise_if_not_found,
                           with_option=True,
                           ))
            byname = None

        ret = []
        for child in self._impl_children:
            if isinstance(child, KernelGroupConfig):
                ret.extend(child.find_group(byname=byname,
                                            bypath=bypath,
                                            byoption=byoption,
                                            byvalue=byvalue,
                                            config_bag=config_bag,
                                            raise_if_not_found=False,
                                            _sub=True))
            else:
                cconfig_bag = config_bag.copy()
                cconfig_bag.context = child
                if cconfig_bag.properties is None:
                    settings = child.get_settings()
                    properties = settings.get_context_properties(child.properties_cache)
                    permissives = settings.get_context_permissives()
                    cconfig_bag.properties = properties
                    cconfig_bag.permissives = permissives
                root_option_bag = OptionBag(child.get_description(),
                                            None,
                                            cconfig_bag,
                                            )
                try:
                    next(child.find(root_option_bag,
                                    None,
                                    byname,
                                    byvalue,
                                    raise_if_not_found=False,
                                    only_path=bypath,
                                    only_option=byoption,
                                    ))
                    ret.append(child)
                except StopIteration:
                    pass
        if not _sub:
            self._find_return_results(ret != [],  # pylint: disable=use-implicit-booleaness-not-comparison
                                      raise_if_not_found,
                                      )
        return ret

    def reset(self,
              path: str,
              config_bag: ConfigBag,
              ) -> None:
        """reset value for specified path
        """
        for child in self._impl_children:
            settings = child.get_settings()
            cconfig_bag = config_bag.copy()
            cconfig_bag.context = child
            settings = child.get_settings()
            properties = settings.get_context_properties(child.properties_cache)
            permissives = settings.get_context_permissives()
            cconfig_bag.properties = properties
            cconfig_bag.permissives = permissives
            cconfig_bag.remove_validation()
            option_bag = child.get_sub_option_bag(cconfig_bag,
                                                  path,
                                                  None,
                                                  False,
                                                  )[-1]
            child.get_values().reset(option_bag)

    def getconfig(self,
                  name: str,
                  ) -> KernelConfig:
        """get a child from a config name
        """
        for child in self._impl_children:
            if name == child.impl_getname():
                return child
        raise ConfigError(_('unknown config "{}"').format(name))


class KernelMixConfig(KernelGroupConfig):
    """Kernel mixconfig: this config can have differents optiondescription tree
    """
    # pylint: disable=too-many-instance-attributes
    __slots__ = ('_impl_symlink',
                 '_storage',
                 )
    impl_type = 'mix'

    def __init__(self,
                 optiondescription,
                 children,
                 name=None,
                 display_name=None,
                 _duplicate=False,
                 ):
        self._impl_name = name
        self._impl_symlink = []
        for child in children:
            if not isinstance(child, (KernelConfig, KernelMixConfig)):
                raise TypeError(_("child must be a Config, MixConfig or MetaConfig"))
            child.parents.append(weakref.ref(self))
        self._impl_settings = Settings()
        self._impl_settings._properties = deepcopy(self._impl_settings._properties)
        self._impl_settings._permissives = deepcopy(self._impl_settings._permissives)
        self._impl_permissives_cache = Cache()
        self.properties_cache = Cache()
        self._impl_values = Values()
        self._impl_values._values = deepcopy(self._impl_values._values)
        self._impl_values_cache = Cache()
        self._display_name = display_name
        self._impl_build_all_caches(optiondescription)
        super().__init__(children,
                         _descr=optiondescription,
                         display_name=display_name,
                         )

    def set_value(self,
                  option_bag,
                  value,
                  only_config=False,
                  force_default=False,
                  force_dont_change_value=False,
                  force_default_if_same=False,
                  ):
        """only_config: could be set if you want modify value in all Config included in
                        this KernelMetaConfig
        """
        # pylint: disable=too-many-branches,too-many-nested-blocks,too-many-locals,too-many-arguments
        ret = []
        if only_config:
            if force_default or force_default_if_same or force_dont_change_value:
                raise ValueError(_('force_default, force_default_if_same or '
                                   'force_dont_change_value cannot be set with'
                                   ' only_config'))
        else:
            if force_default or force_default_if_same or force_dont_change_value:
                if force_default and force_dont_change_value:
                    raise ValueError(_('force_default and force_dont_change_value'
                                       ' cannot be set together'))
                for child in self._impl_children:
                    cconfig_bag = option_bag.config_bag.copy()
                    cconfig_bag.context = child
                    settings = child.get_settings()
                    properties = settings.get_context_properties(child.properties_cache)
                    cconfig_bag.properties = properties
                    cconfig_bag.permissives = settings.get_context_permissives()
                    try:
                        if self.impl_type == 'meta':
                            obj = self
                        else:
                            obj = child
                        validate_properties = not force_default and not force_default_if_same
                        moption_bag = obj.get_sub_option_bag(cconfig_bag,
                                                             option_bag.path,
                                                             option_bag.index,
                                                             validate_properties,
                                                             )[-1]
                        if force_default_if_same:
                            if not child.get_values().hasvalue(option_bag.path):
                                child_value = undefined
                            else:
                                child_value = child.get_value(moption_bag)
                        if force_default or (force_default_if_same and value == child_value):
                            child.get_values().reset(moption_bag)
                            continue
                        if force_dont_change_value:
                            child_value = child.get_value(moption_bag)
                            if value != child_value:
                                child.set_value(moption_bag,
                                                     child_value,
                                                     )
                    except PropertiesOptionError as err:
                        # pylint: disable=protected-access
                        ret.append(PropertiesOptionError(err._option_bag,
                                                         err.proptype,
                                                         err._settings,
                                                         err._opt_type,
                                                         err._name,
                                                         err._orig_opt,
                                                         ))
                    except (ValueError, LeadershipError, AttributeError) as err:
                        ret.append(err)

        try:
            moption_bag = self.get_sub_option_bag(option_bag.config_bag,
                                                  option_bag.path,
                                                  option_bag.index,
                                                  not only_config,
                                                  )[-1]
            if only_config:
                ret = super().set_value(moption_bag,
                                        value,
                                        only_config=only_config,
                                        )
            else:
                _CommonConfig.set_value(self,
                                        moption_bag,
                                        value,
                                        )
        except (PropertiesOptionError, ValueError, LeadershipError) as err:
            ret.append(err)
        return ret

    def reset(self,
              path: str,
              only_children: bool,
              config_bag: ConfigBag,
              ) -> None:
        """reset value for a specified path
        """
        # pylint: disable=arguments-differ
        rconfig_bag = config_bag.copy()
        rconfig_bag.remove_validation()
        if self.impl_type == 'meta':
            option_bag = self.get_sub_option_bag(config_bag,
                                                 path,
                                                 None,
                                                 True,
                                                 )[-1]
        elif not only_children:
            try:
                option_bag = self.get_sub_option_bag(rconfig_bag,
                                                     path,
                                                     None,
                                                     True,
                                                     )[-1]
            except AttributeError:
                only_children = True
        for child in self._impl_children:
            rconfig_bag.context = child
            try:
                if self.impl_type == 'meta':
                    moption_bag = option_bag
                    moption_bag.config_bag = rconfig_bag
                else:
                    moption_bag = child.get_sub_option_bag(rconfig_bag,
                                                           path,
                                                           None,
                                                           True,
                                                           )[-1]
                child.get_values().reset(moption_bag)
            except AttributeError:
                pass
            if isinstance(child, KernelMixConfig):
                child.reset(path,
                            False,
                            rconfig_bag,
                            )
        if not only_children:
            option_bag.config_bag = config_bag
            self.get_values().reset(option_bag)

    def new_config(self,
                   name=None,
                   type_='config',
                   ):
        """Create a new config/metaconfig/mixconfig and add it to this MixConfig"""
        if name:
            for child in self._impl_children:
                if child.impl_getname() == name:
                    raise ConflictError(_('config name must be uniq in '
                                          'groupconfig for {0}').format(child))
        assert type_ in ('config', 'metaconfig', 'mixconfig'), _('unknown type {}').format(type_)
        if type_ == 'config':
            config = KernelConfig(self._impl_descr,
                                  name=name)
        elif type_ == 'metaconfig':
            config = KernelMetaConfig([],
                                      optiondescription=self._impl_descr,
                                      name=name,
                                      )
        elif type_ == 'mixconfig':
            config = KernelMixConfig(children=[],
                                     optiondescription=self._impl_descr,
                                     name=name,
                                     )
        # Copy context properties/permissives
        settings = config.get_settings()
        properties = settings.get_context_properties(config.properties_cache)
        settings.set_context_properties(properties,
                                        config,
                                        )
        settings.set_context_permissives(settings.get_context_permissives())
        settings.ro_append = settings.ro_append
        settings.rw_append = settings.rw_append
        settings.ro_remove = settings.ro_remove
        settings.rw_remove = settings.rw_remove
        settings.default_properties = settings.default_properties

        config.parents.append(weakref.ref(self))
        self._impl_children.append(config)
        return config

    def add_config(self,
                   config,
                   ):
        """Add a child config to a mix config"""
        if not config.impl_getname():
            raise ConfigError(_('config added has no name, the name is mandatory'))
        if config.impl_getname() in [child.impl_getname() for child in self._impl_children]:
            raise ConflictError(_('config name "{0}" is not uniq in '
                                  'groupconfig "{1}"').format(config.impl_getname(),
                                                              self.impl_getname()),
                                )
        config.parents.append(weakref.ref(self))
        self._impl_children.append(config)
        config.reset_cache(None, None)

    def remove_config(self,
                      name,
                      ):
        """Remove a child config to a mix config by it's name"""
        for current_index, child in enumerate(self._impl_children):
            if name == child.impl_getname():
                child.reset_cache(None, None)
                break
        else:
            raise ConfigError(_(f'cannot find the config {name}'))
        for child_index, parent in enumerate(child.parents):
            if parent() == self:
                break
        else:  # pragma: no cover
            raise ConfigError(_('cannot find the config {}').format(self.impl_getname()))
        self._impl_children.pop(current_index)
        child.parents.pop(child_index)
        return child


class KernelMetaConfig(KernelMixConfig):
    """Meta config
    """
    __slots__ = tuple()
    impl_type = 'meta'

    def __init__(self,
                 children,
                 optiondescription=None,
                 name=None,
                 display_name=None,
                 _duplicate=False,
                 ):
        descr = None
        if optiondescription is not None:
            if not _duplicate:
                new_children = []
                for child_name in children:
                    assert isinstance(child_name, str), _('MetaConfig with optiondescription'
                                                          ' must have string has child, '
                                                          'not {}').format(child_name)
                    new_children.append(KernelConfig(optiondescription, name=child_name))
                children = new_children
            descr = optiondescription
        for child in children:
            if __debug__ and not isinstance(child, (KernelConfig,
                                                    KernelMetaConfig)):
                raise TypeError(_("child must be a Config or MetaConfig"))
            if descr is None:
                descr = child.get_description()
            elif descr is not child.get_description():
                raise ValueError(_('all config in metaconfig must '
                                   'have the same optiondescription'))
        super().__init__(descr,
                         children,
                         name=name,
                         display_name=display_name,
                         )

    def add_config(self,
                   config,
                   ):
        if self._impl_descr is not config.get_description():
            raise ValueError(_('metaconfig must '
                               'have the same optiondescription'))
        super().add_config(config)



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""SymLinkOption link to an other option
"""
from typing import Any
from .baseoption import BaseOption, valid_name
from ..error import ConfigError
from ..i18n import _


class SymLinkOption(BaseOption):
    """SymLinkOption link to an other option
    """
    __slots__ = ('_opt',)

    def __init__(self,
                 name: str,
                 opt: BaseOption,
                 ) -> None:
        # pylint: disable=super-init-not-called
        if not valid_name(name):
            raise ValueError(_('"{0}" is an invalid name for an option').format(name))
        if not isinstance(opt, BaseOption) or \
                opt.impl_is_optiondescription() or \
                opt.impl_is_symlinkoption():
            raise ValueError(_('malformed symlinkoption must be an option for symlink {0}'
                               '').format(name))
        _setattr = object.__setattr__
        _setattr(self, '_name', name)
        _setattr(self, '_opt', opt)
        opt._add_dependency(self)
        self._path = None

    def __getattr__(self,
                    name: str,
                    ) -> Any:
        return getattr(self._opt, name)

    def _setsubdyn(self,
                   subdyn,
                   ) -> None:
        raise ConfigError(_('cannot set symlinkoption in a '
                            'dynoptiondescription'))

    def impl_has_dependency(self,
                            self_is_dep: bool=True,
                            ) -> bool:
        """If self_is_dep is True, it has dependency (self._opt), so return True
        if self_is_dep is False, cannot has validation or callback, so return False
        """
        return self_is_dep

    def impl_is_symlinkoption(self) -> bool:
        """it's a symlinkoption
        """
        return True

    def impl_getopt(self) -> BaseOption:
        """get to linked option
        """
        return self._opt

    def issubdyn(self) -> bool:
        """it's not a sub dyn option
        """
        return False

    def impl_is_multi(self) -> bool:
        """is it a multi?
        """
        if self._opt.issubdyn():
            return True
        return self._opt.impl_is_multi()

    def impl_is_submulti(self) -> bool:
        """is it a submulti?
        """
        return self._opt.impl_is_submulti()



# -*- coding: utf-8 -*-
# Copyright (C) 2014-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""OptionDescription
"""
import weakref
from typing import Optional, Iterator, Union, List


from ..i18n import _
from ..setting import ConfigBag, OptionBag, groups, undefined, owners, Undefined
from .baseoption import BaseOption
from .syndynoptiondescription import SubDynOptionDescription, SynDynOptionDescription
from ..error import ConfigError, ConflictError


class CacheOptionDescription(BaseOption):
    """manage cache for option description
    """
    __slots__ = ('_cache_force_store_values',
                 '_cache_dependencies_information',
                 )

    def impl_already_build_caches(self) -> bool:
        """is a readonly option?
        """
        return self.impl_is_readonly()

    def _build_cache(self,
                     _consistencies=None,
                     _consistencies_id=0,
                     currpath: List[str]=None,
                     cache_option=None,
                     force_store_values=None,
                     dependencies_information=None,
                     display_name=None,
                     ) -> None:
        """validate options and set option has readonly option
        """
        # pylint: disable=too-many-branches,too-many-arguments
        # _consistencies is None only when we start to build cache
        if _consistencies is None:
            init = True
            _consistencies = {}
            if __debug__:
                cache_option = []
            force_store_values = []
            dependencies_information = {}
            currpath = []
        else:
            init = False

        if self.impl_is_readonly():
            # cache already set
            raise ConfigError(_('option description seems to be part of an other '
                                'config'))
        for option in self.get_children(config_bag=undefined,  # pylint: disable=no-member
                                        dyn=False,
                                        ):
            if __debug__:
                cache_option.append(option)
            sub_currpath = currpath + [option.impl_getname()]
            subpath = '.'.join(sub_currpath)
            if isinstance(option, OptionDescription):
                # pylint: disable=protected-access
                option._build_cache(_consistencies,
                                    _consistencies_id,
                                    sub_currpath,
                                    cache_option,
                                    force_store_values,
                                    dependencies_information,
                                    display_name,
                                    )
            else:
                for information, options in option.get_dependencies_information().items():
                    if None in options:
                        dependencies_information.setdefault(information, []).append(option)
                if not option.impl_is_symlinkoption():
                    properties = option.impl_getproperties()
                    if 'force_store_value' in properties:
                        force_store_values.append(option)
#                    if __debug__ and ('force_default_on_freeze' in properties or \
#                            'force_metaconfig_on_freeze' in properties) and \
#                            'frozen' not in properties and \
#                            option.impl_is_leader():
#                        raise ConfigError(_('a leader ({0}) cannot have '
#                                            '"force_default_on_freeze" or '
#                                            '"force_metaconfig_on_freeze" '
#                                            'property without "frozen"'
#                                            '').format(option.impl_get_display_name()))
            if option.impl_is_readonly():
                raise ConflictError(_('duplicate option: {0}').format(option))
            if not self.impl_is_readonly() and display_name:
                option._display_name_function = display_name  # pylint: disable=protected-access
            option._path = subpath  # pylint: disable=protected-access
            option._set_readonly()  # pylint: disable=protected-access
        if init:
            self._cache_force_store_values = force_store_values  # pylint: disable=attribute-defined-outside-init
            self._cache_dependencies_information = dependencies_information  # pylint: disable=attribute-defined-outside-init
            self._path = self._name  # pylint: disable=attribute-defined-outside-init,no-member
            self._set_readonly()

    def impl_build_force_store_values(self,
                                      config_bag: ConfigBag,
                                      ) -> None:
        """set value to force_store_values option
        """
        # pylint: disable=too-many-branches
        def do_option_bags(option):
            if option.issubdyn():
                dynopt = option.getsubdyn()
                yield from dynopt.get_sub_children(option,
                                                   config_bag,
                                                   index=None,
                                                   )
            else:
                yield OptionBag(option,
                                None,
                                config_bag,
                                properties=None,
                                )
        if 'force_store_value' not in config_bag.properties:
            return
        values = config_bag.context.get_values()
        for option in self._cache_force_store_values:
            if option.impl_is_follower():
                leader = option.impl_get_leadership().get_leader()
                for leader_option_bag in do_option_bags(leader):
                    leader_option_bag.properties = frozenset()
                    follower_len = len(values.get_value(leader_option_bag)[0])
                    if option.issubdyn():
                        subpath = leader_option_bag.option.rootpath
                        doption = option.to_dynoption(subpath,
                                                      leader_option_bag.option.impl_getsuffix(),
                                                      leader_option_bag.option.dyn_parent,
                                                      )
                    else:
                        doption = option
                    subpath = doption.impl_getpath()
                    for index in range(follower_len):
                        option_bag = OptionBag(doption,
                                               index,
                                               config_bag,
                                               properties=frozenset(),
                                               )
                        if values.hasvalue(subpath, index=index):
                            continue
                        value = values.get_value(option_bag)[0]
                        if value is None:
                            continue
                        values.set_storage_value(subpath,
                                                 index,
                                                 value,
                                                 owners.forced,
                                                 )
            else:
                for option_bag in do_option_bags(option):
                    option_bag.properties = frozenset()
                    value = values.get_value(option_bag)[0]
                    if value is None:
                        continue
                    if values.hasvalue(option_bag.path):
                        continue
                    values.set_storage_value(option_bag.path,
                                             None,
                                             value,
                                             owners.forced,
                                             )


class OptionDescriptionWalk(CacheOptionDescription):
    """get child of option description
    """
    __slots__ = ('_children',)

    def get_child(self,
                  name: str,
                  config_bag: ConfigBag,
                  subpath: str,
                  *,
                  dynoption=None,
                  option_suffix: Optional[str]=None,
                  allow_dynoption: bool=False,
                  ) -> Union[BaseOption, SynDynOptionDescription]:
        """get a child
        """
        # if not dyn
        if name in self._children[0]:  # pylint: disable=no-member
            option = self._children[1][self._children[0].index(name)]  # pylint: disable=no-member
            if option.impl_is_dynoptiondescription():
                if allow_dynoption:
                    option_suffix = None
                else:
                    raise AttributeError(_(f'unknown option "{name}" '
                                           "in root optiondescription (it's a dynamic option)"
                                           ))
            if option.issubdyn():
                return option.to_dynoption(subpath,
                                           option_suffix,
                                           option,
                                           )
            return option
        # if dyn
        if dynoption:
            self_opt = dynoption
        else:
            self_opt = self
        for child in self._children[1]:  # pylint: disable=no-member
            if not child.impl_is_dynoptiondescription():
                continue
            for suffix in child.get_suffixes(config_bag,
                                             dynoption,
                                             ):
                if name != child.impl_getname(suffix):
                    continue
                return child.to_dynoption(subpath,
                                          suffix,
                                          child,
                                          )
        if self.impl_get_group_type() == groups.root:  # pylint: disable=no-member
            raise AttributeError(_(f'unknown option "{name}" '
                                   'in root optiondescription'
                                   ))
        raise AttributeError(_(f'unknown option "{name}" '
                               f'in optiondescription "{self_opt.impl_get_display_name()}"'
                               ))

    def get_children(self,
                     config_bag: Union[ConfigBag, Undefined],
                     *,
                     dyn: bool=True,
                     #path: Optional[str]=None,
                     dynoption=None,
                     option_suffix: Optional[str]=None,
                     ) -> Union[BaseOption, SynDynOptionDescription]:
        """get children
        """
#        if path:
#            subpath = path
        if dynoption:
            self_opt = dynoption
        else:
            self_opt = self
        if not dyn or config_bag is undefined or \
                config_bag.context.get_description() == self:
            subpath = ''
        else:
            subpath = self_opt.impl_getpath()
        for child in self._children[1]:  # pylint: disable=no-member
            if dyn and child.impl_is_dynoptiondescription():
                for suffix in child.get_suffixes(config_bag,
                                                 dynoption,
                                                 ):
                    yield child.to_dynoption(subpath,
                                             suffix,
                                             child,
                                             )
            elif dyn and child.issubdyn() or child.impl_is_dynsymlinkoption():
                yield child.to_dynoption(subpath,
                                         option_suffix,
                                         child,
                                         )
            else:
                yield child

    def get_children_recursively(self,
                                 bytype: Optional[BaseOption],
                                 byname: Optional[str],
                                 config_bag: ConfigBag,
                                 self_opt: BaseOption=None,
                                 ) -> Iterator[Union[BaseOption, SynDynOptionDescription]]:
        """get children recursively
        """
        if self_opt is None:
            self_opt = self
        for option in self_opt.get_children(config_bag):
            if option.impl_is_optiondescription():
                for subopt in option.get_children_recursively(bytype,
                                                              byname,
                                                              config_bag,
                                                              ):
                    yield subopt
            elif (byname is None or option.impl_getname() == byname) and \
                    (bytype is None or isinstance(option, bytype)):
                yield option


class OptionDescription(OptionDescriptionWalk):
    """Config's schema (organisation, group) and container of Options
    The `OptionsDescription` objects lives in the `tiramisu.config.Config`.
    """
    __slots__ = ('_group_type',)

    def __init__(self,
                 name: str,
                 doc: str,
                 children: List[BaseOption],
                 properties=None) -> None:
        """
        :param children: a list of options (including optiondescriptions)

        """
        assert isinstance(children, list), _('children in optiondescription "{}" '
                                             'must be a list').format(name)
        super().__init__(name,
                         doc=doc,
                         properties=properties)
        child_names = []
        if __debug__:
            dynopt_names = []
        for child in children:
            name = child.impl_getname()
            child_names.append(name)
            if __debug__ and child.impl_is_dynoptiondescription():
                dynopt_names.append(name)

        # before sorting
        children_ = (tuple(child_names), tuple(children))

        if __debug__:
            # better performance like this
            child_names.sort()
            old = None
            for child in child_names:
                if child == old:
                    raise ConflictError(_('duplicate option name: '
                                          '"{0}"').format(child))
                if dynopt_names:
                    for dynopt in dynopt_names:
                        if child != dynopt and child.startswith(dynopt):
                            raise ConflictError(_(f'the option\'s name "{child}" start as '
                                                  f'the dynoptiondescription\'s name "{dynopt}"'))
                old = child
        self._children = children_
        # the group_type is useful for filtering OptionDescriptions in a config
        self._group_type = groups.default

    def _setsubdyn(self,
                   subdyn,
                   ) -> None:
        for child in self._children[1]:
            child._setsubdyn(subdyn)
        super()._setsubdyn(subdyn)

    def impl_is_optiondescription(self) -> bool:
        """the option is an option description
        """
        return True

    def impl_is_dynoptiondescription(self) -> bool:
        """the option is not dynamic
        """
        return False

    def impl_is_leadership(self) -> bool:
        """the option is not a leadership
        """
        return False

    # ____________________________________________________________
    def impl_set_group_type(self,
                            group_type: groups.GroupType,
                            ) -> None:
        """sets a given group object to an OptionDescription

        :param group_type: an instance of `GroupType` or `LeadershipGroupType`
                              that lives in `setting.groups`
        """
        if __debug__:
            if self._group_type != groups.default:
                raise ValueError(_('cannot change group_type if already set '
                                   '(old {0}, new {1})').format(self._group_type,
                                                                group_type))
            if not isinstance(group_type, groups.GroupType):
                raise ValueError(_('group_type: {0}'
                                   ' not allowed').format(group_type))
            if isinstance(group_type, groups.LeadershipGroupType):
                raise ConfigError('please use Leadership object instead of OptionDescription')
        self._group_type = group_type

    def impl_get_group_type(self) -> groups.GroupType:
        """get the group type of option description
        """
        return self._group_type

    def to_dynoption(self,
                     rootpath: str,
                     suffix: str,
                     ori_dyn) -> SynDynOptionDescription:
        """get syn dyn option description
        """
        if suffix is None:
            return SubDynOptionDescription(self,
                                           rootpath,
                                           ori_dyn,
                                           )
        return SynDynOptionDescription(self,
                                       rootpath,
                                       suffix,
                                       ori_dyn)

    def impl_is_dynsymlinkoption(self) -> bool:
        """option is not a dyn symlink option
        """
        return False



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
"""IPOption
"""
from ipaddress import ip_address, ip_interface

from ..i18n import _
from .stroption import StrOption


class IPOption(StrOption):
    """represents the choice of an ip
    """
    __slots__ = tuple()
    _type = _('IP')

    def __init__(self,
                 *args,
                 private_only=False,
                 allow_reserved=False,
                 cidr=False,
                 extra=None,
                 **kwargs):
        if extra is None:
            extra = {}
        extra['_private_only'] = private_only
        extra['_allow_reserved'] = allow_reserved
        extra['_cidr'] = cidr
        super().__init__(*args,
                         extra=extra,
                         **kwargs)

    def _validate_cidr(self, value):
        try:
            ip_obj = ip_interface(value)
        except ValueError as err:
            raise ValueError() from err
        if ip_obj.ip == ip_obj.network.network_address:
            raise ValueError(_("it's in fact a network address"))
        if ip_obj.ip == ip_obj.network.broadcast_address:
            raise ValueError(_("it's in fact a broacast address"))

    def _validate_ip(self, value):
        try:
            str(ip_address(value))
        except ValueError as err:
            raise ValueError() from err

    def validate(self,
                 value: str) -> None:
        super().validate(value)
        if self.impl_get_extra('_cidr'):
            if '/' not in value:
                raise ValueError(_('CIDR address must have a "/"'))
            self._validate_cidr(value)
        else:
            self._validate_ip(value)

    def second_level_validation(self,
                                value: str,
                                warnings_only: bool) -> None:
        ip_obj = ip_interface(value)
        if not self.impl_get_extra('_allow_reserved') and ip_obj.is_reserved:
            if warnings_only:
                msg = _("shouldn't be reserved IP")
            else:
                msg = _("mustn't be reserved IP")
            raise ValueError(msg)
        if self.impl_get_extra('_private_only') and not ip_obj.is_private:
            if warnings_only:
                msg = _("should be private IP")
            else:
                msg = _("must be private IP")
            raise ValueError(msg)



# -*- coding: utf-8 -*-
"Leadership support"
# Copyright (C) 2014-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
import weakref
from typing import List, Iterator, Optional


from ..i18n import _
from ..setting import groups, undefined, OptionBag, ALLOWED_LEADER_PROPERTIES
from .optiondescription import OptionDescription
from .syndynoptiondescription import SynDynLeadership
from .baseoption import BaseOption
from .option import Option
from ..error import LeadershipError
from ..autolib import Calculation


class Leadership(OptionDescription):
    """Leadership
    """
    # pylint: disable=too-many-arguments
    __slots__ = ('leader',
                 'followers',
                 )

    def __init__(self,
                 name: str,
                 doc: str,
                 children: List[BaseOption],
                 properties=None) -> None:
        super().__init__(name,
                         doc,
                         children,
                         properties=properties)
        self._group_type = groups.leadership
        followers = []
        if len(children) < 2:
            raise ValueError(_('a leader and a follower are mandatories in leadership "{}"'
                               '').format(name))
        leader = children[0]
        for idx, child in enumerate(children):
            if __debug__:
                self._check_child_is_valid(child)
            if idx != 0:
                if __debug__:
                    self._check_default_value(child)
                # remove empty property for follower
                child._properties = frozenset(child._properties - {'empty', 'unique'})
                followers.append(child)
            child._add_dependency(self)
            child._leadership = weakref.ref(self)
        if __debug__:
            for prop in leader.impl_getproperties():
                if prop not in ALLOWED_LEADER_PROPERTIES and not isinstance(prop, Calculation):
                    raise LeadershipError(_('leader cannot have "{}" property').format(prop))

    def _check_child_is_valid(self, child: BaseOption):
        if child.impl_is_symlinkoption():
            raise ValueError(_('leadership "{0}" shall not have '
                               "a symlinkoption").format(self.impl_get_display_name()))
        if not isinstance(child, Option):
            raise ValueError(_('leadership "{0}" shall not have '
                               'a subgroup').format(self.impl_get_display_name()))
        if not child.impl_is_multi():
            raise ValueError(_('only multi option allowed in leadership "{0}" but option '
                               '"{1}" is not a multi'
                               '').format(self.impl_get_display_name(),
                                          child.impl_get_display_name()))

    def _check_default_value(self, child: BaseOption):
        default = child.impl_getdefault()
        if default != []:
            if child.impl_is_submulti() and isinstance(default, (list, tuple)):
                for val in default:
                    if not isinstance(val, Calculation):
                        calculation = False
                        break
                else:
                    # empty default is valid
                    calculation = True
            else:
                calculation = isinstance(default, Calculation)
            if not calculation:
                raise ValueError(_('not allowed default value for follower option '
                                   f'"{child.impl_get_display_name()}" in leadership '
                                   f'"{self.impl_get_display_name()}"'))

    def _setsubdyn(self,
                   subdyn,
                   ) -> None:
        for chld in self._children[1]:
            chld._setsubdyn(subdyn)
        super()._setsubdyn(subdyn)

    def is_leader(self,
                  opt: Option,
                  ) -> bool:
        """the option is the leader
        """
        leader = self.get_leader()
        return opt == leader or (opt.impl_is_dynsymlinkoption() and opt.opt == leader)

    def get_leader(self) -> Option:
        """get leader
        """
        return self._children[1][0]

    def get_followers(self) -> Iterator[Option]:
        """get all followers
        """
        for follower in self._children[1][1:]:
            yield follower

    def in_same_leadership(self,
                           opt: Option,
                           ) -> bool:
        """check if followers are in same leadership
        """
        if opt.impl_is_dynsymlinkoption():
            opt = opt.opt
        return opt in self._children[1]

    def reset(self, config_bag: 'ConfigBag') -> None:
        """reset follower value
        """
        values = config_bag.context.get_values()
        config_bag = config_bag.copy()
        config_bag.remove_validation()
        for follower in self.get_followers():
            soption_bag = OptionBag(follower,
                                    None,
                                    config_bag,
                                    )
            values.reset(soption_bag)

    def follower_force_store_value(self,
                                   value,
                                   config_bag: 'ConfigBag',
                                   owner,
                                   dyn=None,
                                   ) -> None:
        """apply force_store_value to follower
        """
        if value:
            if dyn is None:
                dyn = self
            values = config_bag.context.get_values()
            for idx, follower in enumerate(dyn.get_children(config_bag)):
                if not idx:
                    # it's a master
                    apply_requires = True
                    indexes = [None]
                else:
                    apply_requires = False
                    indexes = range(len(value))
                foption_bag = OptionBag(follower,
                                        None,
                                        config_bag,
                                        apply_requires=apply_requires,
                                        )
                if 'force_store_value' not in foption_bag.properties:
                    continue
                for index in indexes:
                    foption_bag_index = OptionBag(follower,
                                                  index,
                                                  config_bag,
                                                  )
                    values.set_storage_value(foption_bag_index.path,
                                             index,
                                             values.get_value(foption_bag_index)[0],
                                             owner,
                                             )

    def pop(self,
            index: int,
            config_bag: 'ConfigBag',
            followers: Optional[List[Option]]=undefined,
            ) -> None:
        """pop leader value and follower's one
        """
        if followers is undefined:
            # followers are not undefined only in SynDynLeadership
            followers = self.get_followers()
        config_bag = config_bag.copy()
        config_bag.remove_validation()
        values = config_bag.context.get_values()
        for follower in followers:
            soption_bag = OptionBag(follower,
                                    index,
                                    config_bag,
                                    properties=set(),  # do not check force_default_on_freeze
                                                       # or force_metaconfig_on_freeze
                                    )
            values.reduce_index(soption_bag)

    def reset_cache(self,
                    path: str,
                    config_bag: 'ConfigBag',
                    resetted_opts: List[Option],
                    ) -> None:
        self._reset_cache(path,
                          self.get_leader(),
                          self.get_followers(),
                          config_bag,
                          resetted_opts,
                          )

    def _reset_cache(self,
                     path: str,
                     leader: Option,
                     followers: List[Option],
                     config_bag: 'ConfigBag',
                     resetted_opts: List[Option],
                     ) -> None:
        super().reset_cache(path,
                            config_bag,
                            resetted_opts,
                            )
        leader.reset_cache(leader.impl_getpath(),
                           config_bag,
                           None)
        for follower in followers:
            follower.reset_cache(follower.impl_getpath(),
                                 config_bag,
                                 None,
                                 )

    def impl_is_leadership(self) -> None:
        return True

    def to_dynoption(self,
                     rootpath: str,
                     suffix: str,
                     ori_dyn,
                     ) -> SynDynLeadership:
        return SynDynLeadership(self,
                                rootpath,
                                suffix,
                                ori_dyn,
                                )



# -*- coding: utf-8 -*-
"option types and option description"
# Copyright (C) 2012-2023 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
import warnings
from typing import Any, List, Optional, Dict
from itertools import chain

from .baseoption import BaseOption, submulti
from ..i18n import _
from ..setting import undefined, OptionBag
from ..autolib import Calculation, ParamOption, ParamInformation, ParamSelfInformation
from ..error import ValueWarning, ValueErrorWarning, ValueOptionError
from .syndynoption import SynDynOption


class Option(BaseOption):
    # pylint: disable=too-many-statements,too-many-branches,too-many-arguments,too-many-locals
    """
    Abstract base class for configuration option's.

    Reminder: an Option object is **not** a container for the value.
    """
    __slots__ = ('_extra',
                 '_warnings_only',
                 # multi
                 '_multi',
                 # value
                 '_default',
                 '_default_multi',
                 #
                 '_validators',
                 #
                 '_dependencies_information',
                 '_leadership',
                 '_choice_values',
                 '_choice_values_params',
                )
    _type = None
    def __init__(self,
                 name: str,
                 doc: str,
                 default: Any=undefined,
                 default_multi: Any=None,
                 multi: bool=False,
                 validators: Optional[List[Calculation]]=None,
                 properties: Optional[List[str]]=None,
                 warnings_only: bool=False,
                 extra: Optional[Dict]=None):
        _setattr = object.__setattr__
        if not multi and default_multi is not None:
            raise ValueError(_("default_multi is set whereas multi is False"
                               " in option: {0}").format(name))
        if default is undefined:
            if multi is False:
                default = None
            else:
                default = []
        if multi is True:
            is_multi = True
            _multi = 0
        elif multi is False:
            is_multi = False
            _multi = 1
        elif multi is submulti:
            is_multi = True
            _multi = submulti
        else:
            raise ValueError(_('invalid multi type "{}" for "{}"').format(multi,
                                                                          name,
                                                                          ))
        if _multi != 1:
            _setattr(self, '_multi', _multi)
        if multi is not False and default is None:
            default = []
        super().__init__(name,
                         doc,
                         properties=properties,
                         is_multi=is_multi)
        if validators is not None:
            if __debug__ and not isinstance(validators, list):
                raise ValueError(_(f'validators must be a list of Calculation for "{name}"'))
            for validator in validators:
                if __debug__ and not isinstance(validator, Calculation):
                    raise ValueError(_('validators must be a Calculation for "{}"').format(name))
                self.value_dependency(validator)
                self._validators = tuple(validators)
        if extra is not None and extra != {}:
            _setattr(self, '_extra', extra)
        if warnings_only is True:
            _setattr(self, '_warnings_only', warnings_only)
        if is_multi and default_multi is not None:
            def test_multi_value(value):
                if isinstance(value, Calculation):
                    return
                option_bag = OptionBag(self,
                                       None,
                                       undefined,
                                       properties=None,
                                       )
                try:
                    self.validate(value)
                    self.validate_with_option(value,
                                              option_bag,
                                              loaded=True,
                                              )
                except ValueError as err:
                    str_err = str(err)
                    if not str_err:
                        raise ValueError(_('invalid default_multi value "{0}" '
                                           'for option "{1}"').format(str(value),
                                                                      self.impl_get_display_name())
                                         ) from err
                    raise ValueError(_(f'invalid default_multi value "{value}" for option '
                                       f'"{self.impl_get_display_name()}", {str_err}')
                                     ) from err
            if _multi is submulti:
                if not isinstance(default_multi, Calculation):
                    if not isinstance(default_multi, list):
                        raise ValueError(_('invalid default_multi value "{0}" '
                                           'for option "{1}", must be a list for a submulti'
                                           '').format(str(default_multi),
                                                     self.impl_get_display_name()))
                    for value in default_multi:
                        test_multi_value(value)
            else:
                test_multi_value(default_multi)
            _setattr(self, '_default_multi', default_multi)
        option_bag = OptionBag(self,
                               None,
                               undefined,
                               properties=None,
                               )
        self.impl_validate(default,
                           option_bag,
                           loaded=True,
                           )
        self.impl_validate(default,
                           option_bag,
                           check_error=False,
                           loaded=True,
                           )
        self.value_dependencies(default)
        if (is_multi and default != []) or \
                (not is_multi and default is not None):
            if is_multi and isinstance(default, list):
                default = tuple(default)
            _setattr(self, '_default', default)

    def value_dependencies(self,
                           value: Any,
                           ) -> Any:
        """parse dependancies to add dependencies
        """
        if isinstance(value, list):
            for val in value:
                if isinstance(value, list):
                    self.value_dependencies(val)
                elif isinstance(value, Calculation):
                    self.value_dependency(val)
        elif isinstance(value, Calculation):
            self.value_dependency(value)

    def value_dependency(self,
                         value: Any,
                         ) -> Any:
        """parse dependancy to add dependencies
        """
        for param in chain(value.params.args, value.params.kwargs.values()):
            if isinstance(param, ParamOption):
                # pylint: disable=protected-access
                param.option._add_dependency(self)
                self._has_dependency = True
            elif isinstance(param, ParamInformation):
                dest = self
                if isinstance(param, ParamSelfInformation):
                    opt = self
                elif param.option:
                    dest = param.option
                    opt = self
                else:
                    opt = None
                if not getattr(dest, '_dependencies_information', {}):
                    dest._dependencies_information = {}
                dest._dependencies_information.setdefault(param.information_name, []).append(opt)

    #__________________________________________________________________________
    # option's information

    def impl_is_multi(self) -> bool:
        """is it a multi option
        """
        return getattr(self, '_multi', 1) != 1

    def impl_is_submulti(self) -> bool:
        """is it a submulti option
        """
        return getattr(self, '_multi', 1) == 2

    def impl_is_dynsymlinkoption(self) -> bool:
        """is a dynsymlinkoption?
        """
        return False

    def get_type(self) -> str:
        """get the type of option
        """
        return self._type

    def impl_getdefault(self) -> Any:
        """accessing the default value
        """
        is_multi = self.impl_is_multi()
        default = getattr(self, '_default', undefined)
        if default is undefined:
            if is_multi:
                default = []
            else:
                default = None
        elif is_multi and isinstance(default, tuple):
            default = list(default)
        return default

    def impl_getdefault_multi(self) -> Any:
        """accessing the default value for a multi
        """
        if self.impl_is_submulti():
            default_value = []
        else:
            default_value = None
        return getattr(self, '_default_multi', default_value)

    def impl_get_extra(self,
                       key: str,
                       ) -> Any:
        """if extra parameters are store get it
        """
        extra = getattr(self, '_extra', {})
        if isinstance(extra, tuple):
            if key in extra[0]:
                return extra[1][extra[0].index(key)]
            return None
        return extra.get(key)

    #__________________________________________________________________________
    # validator
    def impl_validate(self,
                      value: Any,
                      option_bag: OptionBag,
                      check_error: bool=True,
                      loaded: bool=False,
                      ) -> bool:
        """Return True if value is really valid
        If not validate or invalid return it returns False
        """
        config_bag = option_bag.config_bag
        force_index = option_bag.index
        is_warnings_only = getattr(self, '_warnings_only', False)
        if check_error and config_bag is not undefined and \
                not 'validator' in config_bag.properties:
            return False

        def _is_not_unique(value, option_bag):
            # if set(value) has not same length than value
            if config_bag is undefined or not check_error or \
                    'unique' not in option_bag.properties:
                return
            lvalue = [val for val in value if val is not None]
            if len(set(lvalue)) == len(lvalue):
                return
            for idx, val in enumerate(value):
                if val not in value[idx+1:]:
                    continue
                raise ValueError(_('the value "{}" is not unique'
                                   '').format(val))

        def calculation_validator(val,
                                  _index,
                                  ):
            for validator in getattr(self, '_validators', []):
                calc_is_warnings_only = hasattr(validator, 'warnings_only') and \
                        validator.warnings_only
                if ((check_error and not calc_is_warnings_only) or
                        (not check_error and calc_is_warnings_only)):
                    try:
                        kwargs = {'allow_value_error': True,
                                  'force_value_warning': calc_is_warnings_only,
                                  }
                        if _index is not None and option_bag.index == _index:
                            soption_bag = option_bag
                        else:
                            soption_bag = option_bag.copy()
                            soption_bag.index = _index
                        kwargs['orig_value'] = value

                        validator.execute(soption_bag,
                                          **kwargs,
                                          )
                    except ValueWarning as warn:
                        warnings.warn_explicit(ValueWarning(val,
                                                            self.get_type(),
                                                            self,
                                                            str(warn),
                                                            _index,
                                                            ),
                                               ValueWarning,
                                               self.__class__.__name__, 319)

        def do_validation(_value,
                          _index,
                          ):
            #
            if isinstance(_value, list):
                raise ValueError(_('which must not be a list').format(_value,
                                                                      self.impl_get_display_name()),
                                 )
            if isinstance(_value, Calculation) and config_bag is undefined:
                return

            if _value is not None:
                if check_error:
                    # option validation
                    self.validate(_value)
                    self.validate_with_option(_value,
                                              option_bag,
                                              loaded=loaded,
                                              )
                if ((check_error and not is_warnings_only) or
                        (not check_error and is_warnings_only)):
                    try:
                        self.second_level_validation(_value,
                                                     is_warnings_only)
                    except ValueError as err:
                        if is_warnings_only:
                            warnings.warn_explicit(ValueWarning(_value,
                                                                self.get_type(),
                                                                self,
                                                                str(err),
                                                                _index),
                                                   ValueWarning,
                                                   self.__class__.__name__, 0)
                        else:
                            raise err
                if not loaded:
                    calculation_validator(_value,
                                          _index,
                                          )
        val = value
        err_index = force_index
        try:
            if not self.impl_is_multi():
                do_validation(val, None)
            elif force_index is not None:
                if self.impl_is_submulti():
                    if not isinstance(value, list):
                        raise ValueError(_('which must be a list'))
                    for val in value:
                        do_validation(val,
                                      force_index,
                                      )
                    _is_not_unique(value,
                                   option_bag,
                                   )
                else:
                    do_validation(val,
                                  force_index,
                                  )
            elif isinstance(value, Calculation) and config_bag is undefined:
                pass
            elif self.impl_is_submulti():
                for err_index, lval in enumerate(value):
                    if isinstance(lval, Calculation):
                        continue
                    if not isinstance(lval, list):
                        raise ValueError(_('which "{}" must be a list of list'
                                           '').format(lval))
                    for val in lval:
                        do_validation(val,
                                            err_index)
                    _is_not_unique(lval, option_bag)
            elif not isinstance(value, list):
                raise ValueError(_('which must be a list'))
            else:
                # FIXME suboptimal, not several time for whole=True!
                for err_index, val in enumerate(value):
                    do_validation(val,
                                  err_index,
                                  )
                _is_not_unique(value, option_bag)
        except ValueError as err:
            if config_bag is undefined or \
                    'demoting_error_warning' not in config_bag.properties:
                raise ValueOptionError(val,
                                       self.get_type(),
                                       option_bag.ori_option,
                                       str(err),
                                       err_index) from err
            warnings.warn_explicit(ValueErrorWarning(val,
                                                     self.get_type(),
                                                     option_bag.ori_option,
                                                     str(err),
                                                     err_index),
                                   ValueErrorWarning,
                                   self.__class__.__name__, 0)
            return False
        return True

    def validate_with_option(self,
                             value: Any,
                             option_bag: OptionBag,
                             loaded: bool,
                             ) -> None:
        """validation function with option
        """

    def second_level_validation(self,
                                value: Any,
                                warnings_only: bool,
                                ) -> None:
        """less import validation function
        """

    def impl_is_leader(self):
        """check if option is a leader in a leadership
        """
        leadership = self.impl_get_leadership()
        if leadership is None:
            return False
        return leadership.is_leader(self)

    def impl_is_follower(self):
        """check if option is a leader in a follower
        """
        leadership = self.impl_get_leadership()
        if leadership is None:
            return False
        return not leadership.is_leader(self)

    def impl_get_leadership(self):
        """get leadership
        """
        leadership = getattr(self, '_leadership', None)
        if leadership is None:
            return leadership
        #pylint: disable=not-callable
        return leadership()

    def to_dynoption(self,
                     rootpath: str,
                     suffix: str,
                     dyn_parent,
                     ) -> SynDynOption:
        """tranforme a dynoption to a syndynoption
        """
        return SynDynOption(self,
                            rootpath,
                            suffix,
                            dyn_parent,
                            )
    def validate(self, value: Any):
        """option needs a validate function
        """
        raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
import os
from tiramisu import __version__


ORI_PACKAGE_NAME = 'tiramisu'
PACKAGE_NAME = os.environ.get('PACKAGE_DST', ORI_PACKAGE_NAME)

if PACKAGE_NAME != ORI_PACKAGE_NAME:
    package_dir = {PACKAGE_NAME: ORI_PACKAGE_NAME}
else:
    package_dir = None

setup(
    version=__version__,
    author="Tiramisu's team",
    author_email='gnunux@gnunux.info',
    name=PACKAGE_NAME,
    description='an options controller tool',
    url='https://framagit.org/tiramisu/tiramisu',
    license='GNU Library or Lesser General Public License (LGPL)',
    provides=['tiramisu_api'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
	    "Natural Language :: English",
        "Natural Language :: French",
    ],
    long_description="""\
An options controller tool
-------------------------------------

Due to more and more available options required to set up an operating system,
compiler options or whatever, it became quite annoying to hand the necessary
options to where they are actually used and even more annoying to add new
options. To circumvent these problems the configuration control was
introduced...

Tiramisu is an options handler and an options controller, wich aims at
producing flexible and fast options access.


This version requires Python 3.5 or later.
""",
    include_package_data=True,
    package_dir=package_dir,
    packages=[PACKAGE_NAME],
)
